import re
from functools import lru_cache
from typing import Any, Awaitable, Callable, Dict, FrozenSet, Iterable, Optional, Union

from blacksheep.baseapp import BaseApplication
from blacksheep.messages import Request, Response
from blacksheep.server.routing import Route, Router
from blacksheep.server.websocket import WebSocket

from .responses import not_found, ok, status_code


class CORSPolicy:
    def __init__(
        self,
        *,
        allow_methods: Union[None, str, Iterable[str]] = None,
        allow_headers: Union[None, str, Iterable[str]] = None,
        allow_origins: Union[None, str, Iterable[str]] = None,
        allow_credentials: bool = False,
        max_age: int = 5,
        expose_headers: Union[None, str, Iterable[str]] = None,
    ) -> None:
        if expose_headers is None:
            expose_headers = self.default_expose_headers()
        self._max_age: int = 300
        self._allow_methods: FrozenSet[str]
        self._allow_headers: FrozenSet[str]
        self._allow_origins: FrozenSet[str]
        self._expose_headers: FrozenSet[str]
        self.allow_methods = allow_methods or []
        self.allow_headers = allow_headers or []
        self.allow_origins = allow_origins or []
        self.allow_credentials = bool(allow_credentials)
        self.expose_headers = expose_headers
        self.max_age = max_age

    def default_expose_headers(self) -> FrozenSet[str]:
        return frozenset(
            value.lower()
            for value in {
                "Transfer-Encoding",
                "Content-Encoding",
                "Vary",
                "Request-Context",
                "Set-Cookie",
                "Server",
                "Date",
            }
        )

    def _normalize_set(
        self, value: Union[None, str, Iterable[str]], ci_function: Callable[[str], str]
    ) -> FrozenSet[str]:
        if value is None:
            return frozenset()
        if isinstance(value, str):
            value = re.split(r"\s|,\s?|;\s?", value)
        return frozenset(map(ci_function, value))

    @property
    def allow_methods(self) -> FrozenSet[str]:
        return self._allow_methods

    @allow_methods.setter
    def allow_methods(self, value) -> None:
        self._allow_methods = self._normalize_set(value, str.upper)

    @property
    def allow_headers(self) -> FrozenSet[str]:
        return self._allow_headers

    @allow_headers.setter
    def allow_headers(self, value) -> None:
        self._allow_headers = self._normalize_set(value, str.lower)

    @property
    def allow_origins(self) -> FrozenSet[str]:
        return self._allow_origins

    @allow_origins.setter
    def allow_origins(self, value) -> None:
        self._allow_origins = self._normalize_set(value, str.lower)

    @property
    def max_age(self) -> int:
        return self._max_age

    @max_age.setter
    def max_age(self, value) -> None:
        int_value = int(value)
        if int_value < 0:
            raise ValueError("max_age must be a positive number")
        self._max_age = int_value

    @property
    def expose_headers(self) -> FrozenSet[str]:
        return self._expose_headers

    @expose_headers.setter
    def expose_headers(self, value) -> None:
        self._expose_headers = self._normalize_set(value, str.lower)

    def allow_any_header(self) -> "CORSPolicy":
        self.allow_headers = frozenset("*")
        return self

    def allow_any_method(self) -> "CORSPolicy":
        self.allow_methods = frozenset("*")
        return self

    def allow_any_origin(self) -> "CORSPolicy":
        self.allow_origins = frozenset("*")
        return self


class CORSConfigurationError(Exception):
    pass


class CORSPolicyNotConfiguredError(CORSConfigurationError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f'The policy with name "{name}" is not configured. '
            "Configure this policy before applying it to request handlers."
        )


class NotRequestHandlerError(CORSConfigurationError):
    def __init__(self) -> None:
        super().__init__(
            "The decorated function is not a request handler. "
            "Apply the @cors() decorator after decorators that define routes."
        )


class CORSStrategy:
    def __init__(self, default_policy: CORSPolicy, router: Router) -> None:
        self.default_policy = default_policy
        self._router = router
        self._policies: Dict[str, CORSPolicy] = {}
        self._policies_by_route: Dict[Route, CORSPolicy] = {}

    @property
    def router(self) -> Router:
        return self._router

    @property
    def policies(self) -> Dict[str, CORSPolicy]:
        return self._policies

    def add_policy(self, name: str, policy: CORSPolicy) -> "CORSStrategy":
        """
        Adds a new CORS policy by name to the overall CORS configuration.

        The CORS policy can then be associated to specific request handlers,
        using the instance of `CORSStrategy` as a function decorator:

        @app.cors("example")
        @app.route("/")
        async def foo():
            ....
        """
        if not name:
            raise CORSConfigurationError(
                "A name is required to configure additional CORS policies."
            )

        if name in self.policies:
            raise CORSConfigurationError(
                f"A policy with name {name} is already configured. "
                "The name of CORS policies must be unique."
            )

        self.policies[name] = policy
        return self

    def get_policy_by_route(self, route: Route) -> Optional[CORSPolicy]:
        return self._policies_by_route.get(route)

    def get_policy_by_route_or_default(self, route: Route) -> CORSPolicy:
        return self.get_policy_by_route(route) or self.default_policy

    def __call__(self, policy: str):
        """Decorates a request handler to bind it to a specific policy by name."""

        def decorator(fn):
            is_match = False
            policy_object = self.policies.get(policy)
            if not policy_object:
                raise CORSPolicyNotConfiguredError(policy)

            for route in self.router:
                if route.handler is fn:
                    self._policies_by_route[route] = policy_object
                    is_match = True

            if not is_match:
                raise NotRequestHandlerError()

            return fn

        return decorator


def _get_cors_error_response(message: str) -> Response:
    response = status_code(400)
    response.add_header(b"CORS-Error", message.encode())
    return response


def _get_invalid_origin_response() -> Response:
    return _get_cors_error_response(
        "The origin of the request is not enabled by CORS rules."
    )


def _get_invalid_method_response() -> Response:
    return _get_cors_error_response(
        "The method of the request is not enabled by CORS rules."
    )


def _get_invalid_header_response(header_name: str) -> Response:
    return _get_cors_error_response(
        f'The "{header_name}" request header is not enabled by CORS rules.'
    )


@lru_cache(maxsize=100)
def _get_encoded_value_for_set(items: FrozenSet[str]) -> bytes:
    if not items:
        return b""
    return ", ".join(items).encode()


@lru_cache(maxsize=20)
def _get_encoded_value_for_max_age(max_age: int) -> bytes:
    return str(max_age).encode()


def get_cors_middleware(
    app: BaseApplication,
    strategy: CORSStrategy,
) -> Callable[[Request, Callable[..., Any]], Awaitable[Response]]:
    async def cors_middleware(request: Request, handler):
        if isinstance(request, WebSocket):
            return await handler(request)

        origin = request.get_first_header(b"Origin")

        if not origin:
            # not a CORS request
            return await handler(request)

        next_request_method = request.get_first_header(b"Access-Control-Request-Method")

        # match policy by route to support route-specific CORS rules,
        # instead of supporting only global CORS rules
        # this approach has the added value that destination routes are validated for
        # OPTIONS requests, instead of assuming a path is handled
        route = strategy.router.get_matching_route(
            next_request_method or request.method, request.url.path
        )

        if route is None:
            return not_found()

        policy = strategy.get_policy_by_route_or_default(route)
        allowed_origins = _get_encoded_value_for_set(policy.allow_origins)
        allowed_methods = _get_encoded_value_for_set(policy.allow_methods)
        expose_headers = _get_encoded_value_for_set(policy.expose_headers)
        max_age = _get_encoded_value_for_max_age(policy.max_age)

        if (
            "*" not in policy.allow_origins
            and origin.decode() not in policy.allow_origins
        ):
            return _get_invalid_origin_response()

        if next_request_method:
            # This is a preflight request;
            if (
                "*" not in policy.allow_methods
                and next_request_method.decode() not in policy.allow_methods
            ):
                return _get_invalid_method_response()

            next_request_headers = request.get_first_header(
                b"Access-Control-Request-Headers"
            )

            if next_request_headers and "*" not in policy.allow_headers:
                for value in next_request_headers.split(b","):
                    str_value = value.strip().decode()
                    if str_value.lower() not in policy.allow_headers:
                        return _get_invalid_header_response(str_value)

            response = ok()
            response.set_header(b"Access-Control-Allow-Methods", allowed_methods)
            response.set_header(b"Access-Control-Allow-Origin", allowed_origins)

            if next_request_headers:
                response.set_header(
                    b"Access-Control-Allow-Headers", next_request_headers
                )

            if policy.allow_credentials:
                response.set_header(b"Access-Control-Allow-Credentials", b"true")

            response.set_header(b"Access-Control-Max-Age", max_age)
            return response

        # regular CORS request (non-preflight)
        if (
            "*" not in policy.allow_methods
            and request.method not in policy.allow_methods
        ):
            return _get_invalid_method_response()

        # CORS response headers must be set even if exceptions are used to
        # control the flow of the application.
        # For example if a request handler throws a "Conflict" exception to handle
        # unique constraints conflicts in a relational database.
        # If CORS response headers weren't set, error details wouldn't be available
        # to the client code, in such circumstances.
        try:
            response = await handler(request)
        except Exception as exc:
            response = await app.handle_request_handler_exception(request, exc)

        response.set_header(b"Access-Control-Allow-Origin", allowed_origins)
        response.set_header(b"Access-Control-Expose-Headers", expose_headers)

        return response

    return cors_middleware



import asyncio
from typing import Any, Dict, List, Optional, Union

MessageType = Union[bytes, Dict[str, Any]]


class MockReceive:
    """
    Class used to mock the messages received by an ASGI framework and passed to the
    web framework.

    Example:

        MockReceive([b'{"error":"access_denied"}'])

    Simulates the ASGI server sending this kind of message:

        {
            "body": b'{"error":"access_denied"}',
            "type": "http.message",
            "more_body": False
        }
    """

    def __init__(self, messages: Optional[List[MessageType]] = None):
        self.messages = messages or []
        self.index = 0

    async def __call__(self):
        try:
            message = self.messages[self.index]
        except IndexError:
            message = b""
        else:
            self.index += 1

        if isinstance(message, dict):
            return message

        await asyncio.sleep(0)
        return {
            "body": message,
            "type": "http.message",
            "more_body": False
            if (len(self.messages) == self.index or not message)
            else True,
        }


class MockSend:
    """
    Class used to mock the `send` calls from an ASGI framework.
    Use this class to inspect the messages sent by the framework.
    """

    def __init__(self):
        self.messages = []

    async def __call__(self, message):
        self.messages.append(message)



from dataclasses import dataclass
from functools import wraps
from typing import Optional

import pytest
from guardpost.authentication import User
from rodi import Services, inject

from blacksheep import Request, Response
from blacksheep.server.application import RequiresServiceContainerError
from blacksheep.server.controllers import APIController, Controller, RoutesRegistry
from blacksheep.server.responses import text
from blacksheep.server.routing import RouteDuplicate
from blacksheep.server.websocket import WebSocket
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from blacksheep.utils import ensure_str
from tests.test_files_serving import get_file_path
from tests.utils.application import FakeApplication


# NB: the following is an example of generic decorator (defined using *args and **kwargs)
# it is used to demonstrate that decorators can be used with normalized methods; however
# functools.@wraps is required,
# so it is the order (custom decorators must appear after router decorators)
def example():
    def example_decorator(fn):
        @wraps(fn)
        async def wrapper(*args, **kwargs):
            return await fn(*args, **kwargs)

        return wrapper

    return example_decorator


@pytest.mark.asyncio
async def test_handler_through_controller(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        def greet(self):
            return "Hello World"

        @get("/")
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

        @get("/foo")
        async def foo(self):
            assert isinstance(self, Home)
            return text("foo")

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"

    await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "foo"


@pytest.mark.asyncio
async def test_ws_handler_through_controller(app):
    app.controllers_router = RoutesRegistry()
    ws = app.controllers_router.ws

    called = False

    class Home(Controller):
        @ws("/web-socket")
        async def foo(self, websocket):
            nonlocal called
            called = True
            assert isinstance(self, Home)
            assert isinstance(websocket, WebSocket)
            await websocket.accept()

    app.setup_controllers()
    await app(
        {"type": "websocket", "path": "/web-socket", "query_string": "", "headers": []},
        MockReceive([{"type": "websocket.connect"}]),
        MockSend(),
    )

    assert called is True


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "path_one,path_two",
    [
        ["/<path:filepath>", "/example/<path:filepath>"],
        ["/{path:filepath}", "/example/{path:filepath}"],
    ],
)
async def test_handler_catch_all_through_controller(path_one, path_two, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        def greet(self):
            return "Hello World"

        @get(path_one)
        async def catch_all(self, filepath: str):
            assert isinstance(self, Home)
            assert isinstance(filepath, str)
            return text(filepath)

        @get(path_two)
        async def catch_all_under_example(self, filepath: str):
            assert isinstance(self, Home)
            assert isinstance(filepath, str)
            return text(f"Example: {filepath}")

        @get("/foo")
        async def foo(self):
            assert isinstance(self, Home)
            return text("foo")

    app.setup_controllers()
    await app(get_example_scope("GET", "/hello.js"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "hello.js"

    await app(
        get_example_scope("GET", "/scripts/a/b/c/hello.js"), MockReceive(), MockSend()
    )

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "scripts/a/b/c/hello.js"

    await app(
        get_example_scope("GET", "/example/a/b/c/hello.js"), MockReceive(), MockSend()
    )

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Example: a/b/c/hello.js"

    await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "foo"


@pytest.mark.asyncio
async def test_handler_through_controller_owned_text_method(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        def greet(self):
            return "Hello World"

        @get("/")
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return self.text(self.greet())

        @get("/foo")
        async def foo(self):
            assert isinstance(self, Home)
            return self.text("foo")

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"

    await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "foo"


@pytest.mark.asyncio
async def test_handler_through_controller_owned_html_method(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        async def index(self):
            assert isinstance(self, Home)
            return self.html(
                """
                <h1>Title</h1>
                <p>Lorem ipsum</p>
                """
            )

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert "<h1>Title</h1>" in body
    assert "<p>Lorem ipsum</p>" in body
    assert app.response.content_type() == b"text/html; charset=utf-8"


@pytest.mark.asyncio
async def test_controller_supports_on_request(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    k = 0

    class Home(Controller):
        def greet(self):
            return "Hello World"

        async def on_request(self, request: Request):
            nonlocal k
            k += 1
            assert isinstance(request, Request)
            assert request.url.path == b"/" if k < 10 else b"/foo"
            return await super().on_request(request)

        @get("/")
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

        @get("/foo")
        async def foo(self):
            assert isinstance(self, Home)
            return text("foo")

    app.setup_controllers()

    for j in range(1, 10):
        await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
        assert app.response.status == 200
        assert k == j

    for j in range(10, 20):
        await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())
        assert app.response.status == 200
        assert k == j


@pytest.mark.asyncio
async def test_controller_supports_on_response(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    k = 0

    class Home(Controller):
        def greet(self):
            return "Hello World"

        async def on_response(self, response: Response):
            nonlocal k
            k += 1
            assert isinstance(response, Response)
            if response.content.body == b"Hello World":
                assert k < 10
            else:
                assert k >= 10
            return await super().on_response(response)

        @get("/")
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

        @get("/foo")
        async def foo(self):
            assert isinstance(self, Home)
            return text("foo")

    app.setup_controllers()

    for j in range(1, 10):
        await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
        assert app.response.status == 200
        assert k == j

    for j in range(10, 20):
        await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())
        assert app.response.status == 200
        assert k == j


@pytest.mark.asyncio
async def test_handler_through_controller_supports_generic_decorator(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        def greet(self):
            return "Hello World"

        @get("/")
        @example()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    body = await app.response.text()
    assert body == "Hello World"
    assert app.response.status == 200


@pytest.mark.asyncio
@pytest.mark.parametrize("value", ["Hello World", "Charlie Brown"])
async def test_controller_with_dependency(value, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Settings:
        def __init__(self, greetings: str):
            self.greetings = greetings

    @inject()
    class Home(Controller):
        def __init__(self, settings: Settings):
            assert isinstance(settings, Settings)
            self.settings = settings

        def greet(self):
            return self.settings.greetings

        @get("/")
        async def index(self, request: Request):
            return text(self.greet())

    app.services.add_instance(Settings(value))

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    body = await app.response.text()
    assert body == value
    assert app.response.status == 200


@pytest.mark.asyncio
@pytest.mark.parametrize("value", ["Hello World", "Charlie Brown"])
async def test_many_controllers(value, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Settings:
        def __init__(self, greetings: str):
            self.greetings = greetings

    @inject()
    class Home(Controller):
        def __init__(self, settings: Settings):
            self.settings = settings

        def greet(self):
            return self.settings.greetings

        @get("/")
        async def index(self, request: Request):
            return text(self.greet())

    class Foo(Controller):
        @get("/foo")
        async def foo(self, request: Request):
            return text("foo")

    app.services.add_instance(Settings(value))

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    body = await app.response.text()
    assert body == value
    assert app.response.status == 200


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "first_pattern,second_pattern",
    [
        ("/", "/"),
        (b"/", b"/"),
        (b"/", "/"),
        ("/", b"/"),
        ("/home", "/home/"),
        (b"/home", b"/home/"),
        ("/home", "/home//"),
        (b"/home", b"/home//"),
        ("/hello/world", "/hello/world/"),
        (b"/hello/world", b"/hello/world//"),
        ("/a/b", "/a/b"),
    ],
)
async def test_controllers_with_duplicate_routes_throw(
    first_pattern, second_pattern, app
):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class A(Controller):
        @get(first_pattern)
        async def index(self, request: Request):
            ...

    class B(Controller):
        @get(second_pattern)
        async def index(self, request: Request):
            ...

    with pytest.raises(RouteDuplicate) as context:
        app.use_controllers()

    error = context.value
    assert "Cannot register route pattern `" + ensure_str(
        first_pattern
    ) + "` for `GET` more than once." in str(error)
    assert (
        "This pattern is already registered for handler "
        "test_controllers_with_duplicate_routes_throw.<locals>.A.index." in str(error)
    )


@pytest.mark.asyncio
async def test_controller_on_request_setting_identity(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        async def on_request(self, request: Request):
            request.identity = User({"id": "001", "name": "Charlie Brown"}, "JWTBearer")

        @get("/")
        async def index(self, request: Request, user: Optional[User]):
            assert hasattr(request, "identity")
            assert isinstance(request.identity, User)
            return text(request.identity["name"])

    app.setup_controllers()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    body = await app.response.text()
    assert body == "Charlie Brown"
    assert app.response.status == 200


@pytest.mark.asyncio
async def test_controller_with_base_route_as_string_attribute(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        route = "/home"

        def greet(self):
            return "Hello World"

        @get()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response.status == 404

    await app(get_example_scope("GET", "/home"), MockReceive(), MockSend())
    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"

    await app(get_example_scope("GET", "/home/"), MockReceive(), MockSend())
    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"


@pytest.mark.asyncio
async def test_application_raises_for_invalid_route_class_attribute(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        route = False

        def greet(self):
            return "Hello World"

        @get()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    with pytest.raises(RuntimeError):
        app.setup_controllers()


@pytest.mark.asyncio
async def test_application_raises_for_controllers_for_invalid_services():
    app = FakeApplication(services=Services())
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        def greet(self):
            return "Hello World"

        @get()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    with pytest.raises(RequiresServiceContainerError):
        app.setup_controllers()


@pytest.mark.asyncio
async def test_controller_with_base_route_as_class_method(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Api(Controller):
        @classmethod
        def route(cls):
            return cls.__name__.lower()

    class Home(Api):
        def greet(self):
            return "Hello World"

        @get()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    class Health(Api):
        @get()
        def alive(self):
            return text("Good")

    app.setup_controllers()
    await app(get_example_scope("GET", "/home"), MockReceive(), MockSend())
    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"

    for value in {"/Health", "/health"}:
        await app(get_example_scope("GET", value), MockReceive(), MockSend())
        assert app.response.status == 200
        body = await app.response.text()
        assert body == "Good"


@pytest.mark.asyncio
async def test_controller_with_base_route_as_class_method_fragments(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Api(Controller):
        @classmethod
        def route(cls):
            return "/api/" + cls.__name__.lower()

    class Home(Api):
        def greet(self):
            return "Hello World"

        @get()
        async def index(self, request: Request):
            assert isinstance(self, Home)
            return text(self.greet())

    class Health(Api):
        @get()
        def alive(self):
            return text("Good")

    app.setup_controllers()
    await app(get_example_scope("GET", "/api/home"), MockReceive(), MockSend())
    assert app.response.status == 200
    body = await app.response.text()
    assert body == "Hello World"

    for value in {"/api/Health", "/api/health"}:
        await app(get_example_scope("GET", value), MockReceive(), MockSend())
        assert app.response.status == 200
        body = await app.response.text()
        assert body == "Good"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "first_pattern,second_pattern", [("/", "/home"), (b"/", b"/home")]
)
async def test_controllers_with_duplicate_routes_with_base_route_throw(
    first_pattern, second_pattern, app
):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    # NB: this test creates ambiguity between the full route of a controller handler,
    # and another handler

    class A(Controller):
        route = "home"

        @get(first_pattern)
        async def index(self, request: Request):
            ...

    class B(Controller):
        @get(second_pattern)
        async def index(self, request: Request):
            ...

    with pytest.raises(RouteDuplicate):
        app.use_controllers()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "first_pattern,second_pattern", [("/", "/home"), (b"/", b"/home")]
)
async def test_controller_with_duplicate_route_with_base_route_throw(
    first_pattern, second_pattern, app
):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    # NB: this test creates ambiguity between the full route of a controller handler,
    # and another handler

    class A(Controller):
        route = "home"

        @get(first_pattern)
        async def index(self, request: Request):
            ...

    @app.route(second_pattern)
    async def home():
        ...

    with pytest.raises(RouteDuplicate):
        app.use_controllers()


@pytest.mark.asyncio
async def test_api_controller_without_version(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get
    post = app.controllers_router.post
    delete = app.controllers_router.delete
    patch = app.controllers_router.patch

    class Cat(APIController):
        @get(":cat_id")
        def get_cat(self, cat_id: str):
            return text("1")

        @patch()
        def update_cat(self):
            return text("2")

        @post()
        def create_cat(self):
            return text("3")

        @delete(":cat_id")
        def delete_cat(self):
            return text("4")

    app.setup_controllers()

    expected_result = {
        ("GET", "/api/cat/100"): "1",
        ("PATCH", "/api/cat"): "2",
        ("POST", "/api/cat"): "3",
        ("DELETE", "/api/cat/100"): "4",
    }

    for key, value in expected_result.items():
        method, pattern = key
        await app(get_example_scope(method, pattern), MockReceive(), MockSend())

        assert app.response.status == 200
        body = await app.response.text()
        assert body == value


@pytest.mark.asyncio
async def test_api_controller_with_version(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get
    post = app.controllers_router.post
    delete = app.controllers_router.delete
    patch = app.controllers_router.patch

    class Cat(APIController):
        @classmethod
        def version(cls) -> Optional[str]:
            return "v1"

        @get(":cat_id")
        def get_cat(self, cat_id: str):
            return text("1")

        @patch()
        def update_cat(self):
            return text("2")

        @post()
        def create_cat(self):
            return text("3")

        @delete(":cat_id")
        def delete_cat(self):
            return text("4")

    app.setup_controllers()

    expected_result = {
        ("GET", "/api/v1/cat/100"): "1",
        ("PATCH", "/api/v1/cat"): "2",
        ("POST", "/api/v1/cat"): "3",
        ("DELETE", "/api/v1/cat/100"): "4",
    }

    for key, value in expected_result.items():
        method, pattern = key
        await app(get_example_scope(method, pattern), MockReceive(), MockSend())

        assert app.response.status == 200
        body = await app.response.text()
        assert body == value


@pytest.mark.asyncio
async def test_api_controller_with_version_2(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get
    post = app.controllers_router.post
    delete = app.controllers_router.delete
    patch = app.controllers_router.patch

    class CatV1(APIController):
        @classmethod
        def version(cls) -> Optional[str]:
            return "v1"

        @get(":cat_id")
        def get_cat(self, cat_id: str):
            return text("1")

        @patch()
        def update_cat(self):
            return text("2")

        @post()
        def create_cat(self):
            return text("3")

        @delete(":cat_id")
        def delete_cat(self):
            return text("4")

    class CatV2(APIController):
        @classmethod
        def version(cls) -> Optional[str]:
            return "v2"

        @get(":cat_id")
        def get_cat(self, cat_id: str):
            return text("5")

        @patch()
        def update_cat(self):
            return text("6")

        @post()
        def create_cat(self):
            return text("7")

        @delete(":cat_id")
        def delete_cat(self):
            return text("8")

    app.setup_controllers()

    expected_result = {
        ("GET", "/api/v1/cat/100"): "1",
        ("PATCH", "/api/v1/cat"): "2",
        ("POST", "/api/v1/cat"): "3",
        ("DELETE", "/api/v1/cat/100"): "4",
        ("GET", "/api/v2/cat/100"): "5",
        ("PATCH", "/api/v2/cat"): "6",
        ("POST", "/api/v2/cat"): "7",
        ("DELETE", "/api/v2/cat/100"): "8",
    }

    for key, value in expected_result.items():
        method, pattern = key
        await app(get_example_scope(method, pattern), MockReceive(), MockSend())

        assert app.response.status == 200
        body = await app.response.text()
        assert body == value


@pytest.mark.asyncio
async def test_controller_parameter_name_match(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Example(Controller):
        @get("/")
        async def from_query(self, example: str):
            assert isinstance(self, Example)
            assert isinstance(example, str)
            return text(example)

        @get("/{example}")
        async def from_route(self, example: str):
            assert isinstance(self, Example)
            assert isinstance(example, str)
            return text(example)

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 400
    body = await app.response.text()
    assert body == "Bad Request: Missing query parameter `example`"

    await app(get_example_scope("GET", "/foo"), MockReceive(), MockSend())

    assert app.response.status == 200
    body = await app.response.text()
    assert body == "foo"


@pytest.mark.asyncio
async def test_controller_return_file(app):
    file_path = get_file_path("example.config", "files2")

    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Example(Controller):
        @get("/")
        async def home(self):
            return self.file(file_path, "text/plain; charset=utf-8")

    app.setup_controllers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/plain; charset=utf-8"
    assert response.headers.get_single(b"content-disposition") == b"attachment"

    text = await response.text()
    with open(file_path, mode="rt", encoding="utf8") as f:
        contents = f.read()
        assert contents == text


@dataclass
class Foo:
    name: str
    value: float


@pytest.mark.asyncio
async def test_handler_through_controller_default_type(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        async def index(self) -> Foo:
            return Foo("Hello", 5.5)

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200
    data = await app.response.json()
    assert data == {"name": "Hello", "value": 5.5}


@pytest.mark.asyncio
async def test_handler_through_controller_default_str(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        async def index(self) -> str:
            return "Hello World"

    app.setup_controllers()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200
    data = await app.response.text()
    assert data == "Hello World"



from blacksheep.messages import Request, Response


def _get_hsts_value(max_age: int, include_subdomains: bool) -> bytes:
    value = f"max-age={max_age};"

    if include_subdomains:
        value = value + " includeSubDomains;"

    return value.encode()


class HSTSMiddleware:
    """
    Middleware configuring "Strict-Transport-Security" header on responses.
    By default, it uses "max-age=31536000; includeSubDomains;".
    """

    def __init__(
        self,
        max_age: int = 31536000,
        include_subdomains: bool = True,
    ) -> None:
        self._value = _get_hsts_value(max_age, include_subdomains)

    async def __call__(self, request: Request, handler):
        response: Response = await handler(request)
        response.headers.add(b"Strict-Transport-Security", self._value)
        return response



from datetime import datetime
from typing import Any

import pytest
from guardpost import Identity

from blacksheep.messages import Request, Response
from blacksheep.server.authentication.cookie import CookieAuthentication
from blacksheep.server.dataprotection import generate_secret


def get_auth_cookie(handler: CookieAuthentication, data: Any) -> str:
    response = Response(200)
    handler.set_cookie(data, response)
    return handler.cookie_name + "=" + response.cookies[handler.cookie_name].value


@pytest.mark.asyncio
async def test_cookie_authentication():
    handler = CookieAuthentication()

    request = Request("GET", b"/", headers=[])

    await handler.authenticate(request)

    assert request.identity is not None
    assert request.identity.is_authenticated() is False

    request = Request(
        "GET",
        b"/",
        headers=[
            (
                b"cookie",
                get_auth_cookie(
                    handler, {"id": 1, "email": "example@neoteroi.dev"}
                ).encode(),
            )
        ],
    )

    await handler.authenticate(request)

    assert isinstance(request.identity, Identity)
    assert request.identity.is_authenticated() is True
    assert request.identity.authentication_mode == handler.auth_scheme
    assert request.identity.claims.get("email") == "example@neoteroi.dev"


@pytest.mark.asyncio
async def test_cookie_authentication_handles_invalid_signature():
    handler = CookieAuthentication()

    request = Request(
        "GET",
        b"/",
        headers=[
            (
                b"cookie",
                get_auth_cookie(
                    handler, {"id": 1, "email": "example@neoteroi.dev"}
                ).encode(),
            )
        ],
    )

    other_handler = CookieAuthentication(secret_keys=[generate_secret()])
    await other_handler.authenticate(request)

    assert request.identity is not None
    assert request.identity.is_authenticated() is False


def test_cookie_authentication_unset_cookie():
    handler = CookieAuthentication()

    response = Response(200)
    handler.unset_cookie(response)

    cookie_header = response.cookies[handler.cookie_name]
    assert cookie_header is not None
    assert cookie_header.expires is not None
    assert cookie_header.expires < datetime.utcnow()



from abc import ABC, abstractmethod
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network, ip_address
from typing import AnyStr, Iterable, List, Optional, Sequence, Union

from blacksheep.exceptions import BadRequest
from blacksheep.headers import Headers
from blacksheep.messages import Request

from .hosts import TrustedHostsMiddleware

IPAddress = Union[IPv4Address, IPv6Address]
IPNetwork = Union[IPv4Network, IPv6Network]


class TooManyHeaders(BadRequest):
    def __init__(self, header_name: bytes):
        super().__init__(f"Too many {header_name.decode()} headers.")
        self.header_name = header_name.decode()


class TooManyForwardValues(BadRequest):
    def __init__(self, count: int):
        super().__init__(f"Too many forward values: {count}.")


class InvalidProxyIPError(BadRequest):
    def __init__(self, proxy_ip: IPAddress):
        super().__init__(f"Proxy IP not recognized: {proxy_ip}")


class BaseForwardedHeadersMiddleware(TrustedHostsMiddleware, ABC):
    def __init__(
        self,
        allowed_hosts: Optional[Sequence[str]] = None,
        known_proxies: Optional[Sequence[IPAddress]] = None,
        known_networks: Optional[Sequence[IPNetwork]] = None,
        forward_limit: int = 1,
        accept_only_proxied_requests: bool = True,
    ) -> None:
        super().__init__(allowed_hosts)
        self.known_proxies: List[IPAddress] = (
            list(known_proxies) if known_proxies else [ip_address("127.0.0.1")]
        )
        self.known_networks: List[IPNetwork] = (
            list(known_networks) if known_networks else []
        )
        self.accept_only_proxied_requests = accept_only_proxied_requests
        self.forward_limit = forward_limit

    def parse_ip(self, value: str) -> IPAddress:
        return ip_address(value)

    def should_validate_client_ip(self) -> bool:
        """
        Returns a value indicating whether the request client ip should be validated.

        If known proxies or known networks are configured, and
        `accept_only_proxied_requests` is enabled, then the web framework accepts only
        requests that are proxied.

        If `accept_only_proxied_requests` is set to False, it means the server will
        acceptd both requests that are proxied and requests that are hitting the web
        server directly.
        """
        return (
            self.accept_only_proxied_requests
            and any(self.known_proxies)
            or any(self.known_networks)
        )

    def validate_proxy_ip(self, proxy_ip: IPAddress) -> None:
        if self.known_proxies:
            if proxy_ip in self.known_proxies:
                return

        if self.known_networks:
            if any(network for network in self.known_networks if proxy_ip in network):
                return

        raise InvalidProxyIPError(proxy_ip)

    def validate_proxies_ips(self, proxies: List[IPAddress]) -> None:
        for proxy_ip in proxies:
            self.validate_proxy_ip(proxy_ip)

    @abstractmethod
    async def __call__(self, request: Request, handler):
        """Middleware callback."""


class ForwardedHeaderEntry:
    def __init__(
        self,
        forwarded_for: str,
        forwarded_by: str = "",
        forwarded_host: str = "",
        forwarded_proto: str = "",
    ):
        self.forwarded_for = forwarded_for
        self.forwarded_by = forwarded_by
        self.forwarded_host = forwarded_host
        self.forwarded_proto = forwarded_proto

    def __eq__(self, o: object) -> bool:
        if isinstance(o, ForwardedHeaderEntry):
            return self.__dict__ == o.__dict__
        if isinstance(o, dict):
            return self.__dict__ == o
        return False


def _strip_chars(value):
    return value.strip('"[').replace("]", "")


def parse_forwarded_header(header_value: AnyStr) -> Iterable[ForwardedHeaderEntry]:
    if isinstance(header_value, bytes):
        value = header_value.decode()
    else:
        value = header_value

    groups = value.split(",")

    for group in groups:
        group = group.strip()
        directives = group.split(";")

        forwarded_for: str = ""
        forwarded_by: str = ""
        forwarded_host: str = ""
        forwarded_proto: str = ""

        for directive in directives:
            directive = directive.strip()

            if directive.lower().startswith("for="):
                forwarded_for = _strip_chars(directive[4:])

            if directive.lower().startswith("by="):
                forwarded_by = _strip_chars(directive[3:])

            if directive.lower().startswith("host="):
                forwarded_host = _strip_chars(directive[5:])

            if directive.lower().startswith("proto="):
                forwarded_proto = _strip_chars(directive[6:])

        yield ForwardedHeaderEntry(
            forwarded_for, forwarded_by, forwarded_host, forwarded_proto
        )


class ForwardedHeadersMiddleware(BaseForwardedHeadersMiddleware):
    """
    Class that handles the standard Forwarded header. This middleware should be
    configured early in the chain of middlewares, as it validates and updates each
    request to apply the proper protocol, host, and client ip information.
    """

    def __init__(
        self,
        allowed_hosts: Optional[Sequence[str]] = None,
        known_proxies: Optional[Sequence[IPAddress]] = None,
        known_networks: Optional[Sequence[IPNetwork]] = None,
        forward_limit: int = 1,
        accept_only_proxied_requests: bool = True,
    ) -> None:
        super().__init__(
            allowed_hosts,
            known_proxies,
            known_networks,
            forward_limit=forward_limit,
            accept_only_proxied_requests=accept_only_proxied_requests,
        )

    def get_forwarded_values(self, headers: Headers) -> Iterable[ForwardedHeaderEntry]:
        forwarded_headers = headers[b"Forwarded"]

        for header in forwarded_headers:
            yield from parse_forwarded_header(header)

    def validate_forwarded_entries(
        self, entries: Sequence[ForwardedHeaderEntry]
    ) -> None:
        if len(entries) > self.forward_limit:
            raise TooManyForwardValues(len(entries))

        for entry in entries:
            if entry.forwarded_host:
                self.validate_host(entry.forwarded_host)

            if entry.forwarded_by:
                # the value can be an IP address, but it does not have to;
                # it can be any string
                # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Forwarded
                try:
                    ip = ip_address(entry.forwarded_by)
                except ValueError:
                    pass
                else:
                    self.validate_proxy_ip(ip)

    async def __call__(self, request: Request, handler):
        # Forwarded: by=<identifier>;for=<identifier>;host=<host>;proto=<http|https>
        if self.should_validate_client_ip():
            self.validate_proxy_ip(ip_address(request.client_ip))

        forwarded_entries = list(self.get_forwarded_values(request.headers))

        if forwarded_entries:
            self.validate_forwarded_entries(forwarded_entries)
            first_entry = forwarded_entries[0]

            if first_entry.forwarded_host:
                request.host = first_entry.forwarded_host

            if first_entry.forwarded_proto:
                request.scheme = first_entry.forwarded_proto

            if first_entry.forwarded_for:
                # Note: if the proxy server is configured to send an obfuscated
                # identifier, like "_hidden", we set it here anyway
                request.original_client_ip = first_entry.forwarded_for

        return await handler(request)


class XForwardedHeadersMiddleware(BaseForwardedHeadersMiddleware):
    """
    Class that handles the de-facto standard X-Forwarded headers. This middleware should
    be configured early in the chain of middlewares, as it validates and updates each
    request to apply the proper protocol, host, and client ip information.
    """

    def __init__(
        self,
        allowed_hosts: Optional[Sequence[str]] = None,
        known_proxies: Optional[Sequence[IPAddress]] = None,
        known_networks: Optional[Sequence[IPNetwork]] = None,
        forward_limit: int = 1,
        accept_only_proxied_requests: bool = True,
    ) -> None:
        """
        Creates a new instance of XForwardedHeadersMiddleware which handles X-Forwarded
        headers. By default only 1 forward level is enabled (this can be increased using
        `forward_limit` option. It is possible to change the names of the headers
        that are handled by this class (by default X-Forwarded-For/Host/Proto).
        Requests are updated by this middleware to expose information about the original
        client, which is useful in some circumstances, such as recreating the original
        URL of the request.

        Parameters
        ----------
        allowed_hosts : Optional[Sequence[str]], optional
            If provided, enables validation of the request's direct host value and of
            forwarded-host header values, by default None
        known_proxies : Optional[Sequence[IPAddress]], optional
            If provided, enables validation of proxy IP addresses through IP list
            configuration, by default None
        known_networks : Optional[Sequence[IPNetwork]], optional
            If provided, enables validation of proxy IP addresses through networks
            configuration, by default None
        forward_limit : int, optional
            The maximum number of forwards that are allowed. By default 1
        """
        super().__init__(
            allowed_hosts,
            known_proxies,
            known_networks,
            forward_limit=forward_limit,
            accept_only_proxied_requests=accept_only_proxied_requests,
        )
        self.forwarded_for_header_name = b"X-Forwarded-For"
        self.forwarded_host_header_name = b"X-Forwarded-Host"
        self.forwarded_proto_header_name = b"X-Forwarded-Proto"

    def get_forwarded_for(self, headers: Headers) -> List[IPAddress]:
        # X-Forwarded-For: <client>, <proxy1>, <proxy2>
        forwarded_for_headers: List[bytes] = list(
            headers[self.forwarded_for_header_name]
        )

        if not forwarded_for_headers:
            return []

        if len(forwarded_for_headers) > 1:
            raise TooManyHeaders(self.forwarded_for_header_name)

        forwarded_for = forwarded_for_headers[0].decode().split(",")
        return [
            self.parse_ip(addr) for addr in (a.strip() for a in forwarded_for) if addr
        ]

    def get_forwarded_proto(self, headers: Headers) -> List[str]:
        # X-Forwarded-Proto: https
        forwarded_proto_headers: List[bytes] = list(
            headers[self.forwarded_proto_header_name]
        )

        if not forwarded_proto_headers:
            return []

        if len(forwarded_proto_headers) > 1:
            raise TooManyHeaders(self.forwarded_proto_header_name)

        forwarded_proto = forwarded_proto_headers[0].decode().split(",")
        return [p.strip() for p in forwarded_proto]

    def get_forwarded_host(self, headers: Headers) -> Optional[str]:
        # X-Forwarded-Host: id42.example-cdn.com
        # original host requested by the client in the Host HTTP request header.
        forwarded_hosts = headers[self.forwarded_host_header_name]

        if not forwarded_hosts:
            return None

        if len(forwarded_hosts) > 1:
            raise TooManyHeaders(self.forwarded_host_header_name)

        return forwarded_hosts[0].decode()

    def validate_forwarded_for(self, values: List[IPAddress]) -> None:
        if len(values) > self.forward_limit:
            raise TooManyForwardValues(len(values))

        # the first value is the ID of the client - that is valid
        _, *proxies = values

        self.validate_proxies_ips(proxies)

    def validate_forwarded_proto(self, values: List[str]) -> None:
        if len(values) > self.forward_limit:
            raise TooManyForwardValues(len(values))

    async def __call__(self, request: Request, handler):
        if self.should_validate_client_ip():
            self.validate_proxy_ip(ip_address(request.client_ip))

        headers = request.headers
        forwarded_for = self.get_forwarded_for(headers)
        forwarded_host = self.get_forwarded_host(headers)
        forwarded_proto = self.get_forwarded_proto(headers)

        if forwarded_host:
            self.validate_host(forwarded_host)

            request.host = forwarded_host
        else:
            # validate the host from the Host header anyway
            self.validate_host(request.host)

        if forwarded_for:
            self.validate_forwarded_for(forwarded_for)

            request.original_client_ip = str(forwarded_for[0])

        if forwarded_proto:
            self.validate_forwarded_proto(forwarded_proto)

            request.scheme = forwarded_proto[0]

        return await handler(request)



import pytest

from blacksheep.multipart import (
    FormPart,
    _remove_last_crlf,
    parse_multipart,
    parse_part,
)

from .examples.multipart import (
    FIELDS_THREE_VALUES,
    FIELDS_WITH_CARRIAGE_RETURNS,
    FIELDS_WITH_CARRIAGE_RETURNS_AND_DEFAULT_CHARSET,
    FIELDS_WITH_SMALL_PICTURE,
)


@pytest.mark.parametrize(
    "value,expected_value",
    [
        [
            FIELDS_THREE_VALUES,
            [
                FormPart(b"one", b"aaa"),
                FormPart(b"one", b"bbb"),
                FormPart(b"two", b"ccc"),
                FormPart(b"two", b"daup"),
                FormPart(b"Submit", b"Submit"),
            ],
        ],
        [
            FIELDS_WITH_CARRIAGE_RETURNS,
            [
                FormPart(b"one", b"AA"),
                FormPart(b"one", b"BB"),
                FormPart(b"two", b"CC"),
                FormPart(b"two", b"DD"),
                FormPart(
                    b"description",
                    "Hello\r\n\r\nThis contains øø\r\n\r\nCarriage returns".encode(
                        "utf8"
                    ),
                ),
                FormPart(b"Submit", b"Submit"),
            ],
        ],
        [
            FIELDS_WITH_CARRIAGE_RETURNS_AND_DEFAULT_CHARSET,
            [
                FormPart(b"one", b"AA"),
                FormPart(b"two", b"CC", charset=b"iso-8859-1"),
                FormPart(b"two", b"DD", charset=b"iso-8859-1"),
                FormPart(
                    b"description",
                    "Hello\r\n\r\nThis contains øø\r\n\r\nCarriage returns".encode(
                        "iso-8859-1"
                    ),
                    charset=b"iso-8859-1",
                ),
                FormPart(
                    b"album", "ØØ Void".encode("iso-8859-1"), charset=b"iso-8859-1"
                ),
            ],
        ],
        [
            FIELDS_WITH_SMALL_PICTURE,
            [
                FormPart(b"one", b"aaa"),
                FormPart(b"one", b"bbb"),
                FormPart(b"two", b"ccc"),
                FormPart(b"two", b"example"),
                FormPart(
                    b"file_example",
                    b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x08\x00\x00\x00\x08"
                    b"\x08\x06\x00\x00\x00\xc4\x0f\xbe\x8b\x00\x00\x01\x84iCCPICC profile\x00\x00("
                    b"\x91}\x91=H\xc3@\x1c\xc5_S\xa5R\xaa\x0ev\x90\xe2\x90\xa1:Y\x10\x15q\x94*\x16\xc1Bi+\xb4"
                    b"\xea`r\xe9\x174iHR\\\x1c\x05\xd7\x82\x83\x1f\x8bU\x07\x17g]\x1d\\\x05A\xf0\x03\xc4\xc5"
                    b"\xd5I\xd1EJ\xfc_Rh\x11\xe3\xc1q?\xde\xdd{"
                    b'\xdc\xbd\x03\x84f\x95\xa9f\xcf\x04\xa0j\x96\x91N\xc4\xc5\\~U\x0c\xbc"\x08?\x06\x10'
                    b"\x81Ob\xa6\x9e\xcc,"
                    b"f\xe19\xbe\xee\xe1\xe3\xeb]\x8cgy\x9f\xfbs\xf4+\x05\x93\x01>\x91x\x8e\xe9\x86E\xbcA"
                    b"<\xb3i\xe9\x9c\xf7\x89\xc3\xac,"
                    b")\xc4\xe7\xc4\xe3\x06]\x90\xf8\x91\xeb\xb2\xcbo\x9cK\x0e\x0b<3ld\xd3\xf3\xc4ab\xb1\xd4"
                    b"\xc5r\x17\xb3\xb2\xa1\x12O\x13G\x15U\xa3|!\xe7\xb2\xc2y\x8b\xb3Z\xad\xb3\xf6=\xf9\x0bC"
                    b"\x05m%\xc3u\x9a#H`\tI\xa4 "
                    b'BF\x1d\x15Ta!F\xabF\x8a\x894\xed\xc7=\xfc\x11\xc7\x9f"\x97L\xae\n\x189\x16P\x83\n\xc9'
                    b"\xf1\x83\xff\xc1\xefn\xcd\xe2\xd4\xa4\x9b\x14\x8a\x03\xbd/\xb6\xfd1\n\x04v\x81V\xc3\xb6"
                    b"\xbf\x8fm\xbbu\x02\xf8\x9f\x81+\xad\xe3\xaf5\x81\xd9O\xd2\x1b\x1d-z\x04\x0cn\x03\x17"
                    b"\xd7\x1dM\xde\x03.w\x80\xe1']2$G\xf2\xd3\x14\x8aE\xe0\xfd\x8c\xbe)\x0f\x0c\xdd\x02"
                    b"\xc15\xb7\xb7\xf6>N\x1f\x80,"
                    b"u\xb5|\x03\x1c\x1c\x02c%\xca^\xf7xw_wo\xff\x9ei\xf7\xf7\x03\x0e\xd5r\x7fYl\xff3\x00\x00"
                    b'\x00\x06bKGD\x00\x03\x00\x1a\x00\x1f"Z\xea\xfc\x00\x00\x00\tpHYs\x00\x00.#\x00\x00'
                    b".#\x01x\xa5?v\x00\x00\x00\x07tIME\x07\xe3\n\x03\x0e\x1e\x0fJa{"
                    b"\x87\x00\x00\x00\x19tEXtComment\x00Created with "
                    b"GIMPW\x81\x0e\x17\x00\x00\x00nIDAT\x18\xd3c0\xdan\xf4?.\x98\xe1\xff\x83`\x86\xff\x01"
                    b".+\xff/\x98\xf6\xfc\xff\xcd7~\xff\x95["
                    b"\x8b\xff\xd7\xbd\xb8\xfa\x9f\x01\x9f$\xb3\x94\xfc\x7f\x06|\x92\xccR\xf2\xff\x19\xf0I2K"
                    b"\xc9\xffg\xc0'\xc9,"
                    b"%\xff\x9f\xf1\xe6\x1b\xbf\xff^3U\x19\xa2\x93\x93\x18Z\x8d\xbc\x18\xd0\x01\x13>I\x06\x06"
                    b"\x06\x06\xc6\xba\x17W\xff\xe3\x92d```\x00\x00+\x05h|\x7f\xbaa\x83\x00\x00\x00\x00IEND"
                    b"\xaeB`\x82",
                    b"image/png",
                    b"example-001.png",
                    None,
                ),
                FormPart(b"description", b"Beee\r\n\r\nBeeeeee!"),
                FormPart(b"Submit", b"Submit"),
            ],
        ],
    ],
)
def test_function(value: bytes, expected_value):
    values = list(parse_multipart(value))
    assert values == expected_value


@pytest.mark.parametrize(
    "input,output",
    [
        (b"example", b"example"),
        (b"example\r\n", b"example"),
        (b"example\n", b"example"),
        (b"example\r\n\r\n", b"example\r\n"),
        (b"example\n\n", b"example\n"),
    ],
)
def test_remove_last_crlf(input, output):
    assert _remove_last_crlf(input) == output


def test_parse_part_raises_for_missing_content_disposition():
    with pytest.raises(ValueError):
        parse_part(
            b'X-Content: form-data; name="one"\r\n\r\naaa',
            None,
        )



import pytest

from blacksheep.ranges import InvalidRangeValue, Range, RangePart


@pytest.mark.parametrize(
    "range_a,range_b,equals",
    [
        [
            Range("bytes", [RangePart(100, None)]),
            Range("bytes", [RangePart(100, None)]),
            True,
        ],
        [
            Range("bytes", [RangePart(100, None)]),
            Range("bytes", [RangePart(None, 100)]),
            False,
        ],
        [
            Range("bytes", [RangePart(100, 200), RangePart(400, 600)]),
            Range("bytes", [RangePart(100, 200), RangePart(400, 600)]),
            True,
        ],
        [
            Range("bytes", [RangePart(100, 200), RangePart(400, 600)]),
            Range("bytes", [RangePart(100, 200), RangePart(400, 650)]),
            False,
        ],
        [
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
            True,
        ],
    ],
)
def test_range_equality(range_a, range_b, equals):
    assert (range_a == range_b) is equals


@pytest.mark.parametrize(
    "value,expected_value",
    [
        [Range("bytes", [RangePart(100, None)]), False],
        [Range("bytes", [RangePart(100, 200), RangePart(400, 600)]), True],
        [
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
            True,
        ],
    ],
)
def test_range_is_multipart(value: Range, expected_value: bool):
    assert value.is_multipart == expected_value


@pytest.mark.parametrize(
    "value,expected_range",
    [
        (b"bytes=100-", Range("bytes", [RangePart(100, None)])),
        (b"bytes=-120", Range("bytes", [RangePart(None, 120)])),
        ("bytes=-120", Range("bytes", [RangePart(None, 120)])),
        (
            "bytes=100-200, 400-600, 300-500",
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
        ),
        (
            "bytes=100-200,400-600,300-500",
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
        ),
        (
            "bytes=1200-, -100",
            Range("bytes", [RangePart(1200, None), RangePart(None, 100)]),
        ),
    ],
)
def test_parse_range(value, expected_range):
    parsed_range = Range.parse(value)
    assert parsed_range == expected_range


@pytest.mark.parametrize(
    "value,item",
    [
        ("bytes=100-", Range("bytes", [RangePart(100, None)])),
        ("bytes=-120", Range("bytes", [RangePart(None, 120)])),
        ("bytes=-120", Range("bytes", [RangePart(None, 120)])),
        (
            "bytes=100-200, 400-600, 300-500",
            Range(
                "bytes", [RangePart(100, 200), RangePart(400, 600), RangePart(300, 500)]
            ),
        ),
        (
            "bytes=1200-, -100",
            Range("bytes", [RangePart(1200, None), RangePart(None, 100)]),
        ),
    ],
)
def test_range_repr(value, item):
    assert repr(item) == f"<Range {value}>"


@pytest.mark.parametrize(
    "invalid_value",
    [
        "1000",
        "100-200",
        "bytes=10",
        "bytes=100--200",
        "bytes=100AA, 400-600, 300-500",
        "bytes=100-200, 400-600, 300-A0",
        "bytes=100-AA, 400-600, 300-500",
    ],
)
def test_raise_for_invalid_value(invalid_value):
    with pytest.raises(InvalidRangeValue):
        Range.parse(invalid_value)


@pytest.mark.parametrize(
    "item", [Range("bytes", [RangePart(100, None)]), RangePart(100, None)]
)
def test_range_eq_not_implemented(item):
    value = item.__eq__(True)
    assert value is NotImplemented

    value = item is True
    assert value is False

    value = item == 100
    assert value is False


def test_range_part_raises_if_start_gt_end():
    with pytest.raises(ValueError):
        RangePart(400, 300)

    with pytest.raises(ValueError):
        part = RangePart(100, 300)
        part.start = 400

    with pytest.raises(ValueError):
        part = RangePart(100, 300)
        part.end = 50


def test_range_part_raises_if_any_part_is_negative():
    with pytest.raises(ValueError):
        RangePart(-100, 0)

    with pytest.raises(ValueError):
        part = RangePart(100, 300)
        part.start = -100

    with pytest.raises(ValueError):
        part = RangePart(100, 300)
        part.end = -50


def test_range_part_can_satisfy_raises_if_both_ends_are_none():
    part = RangePart(None, None)

    with pytest.raises(TypeError):
        part.can_satisfy(200)


def test_range_part_helper_methods():
    part_without_start = RangePart(None, 300)
    part_without_end = RangePart(500, None)

    assert part_without_start.is_suffix_length is True
    assert part_without_start.is_to_end is False

    assert part_without_end.is_to_end is True
    assert part_without_end.is_suffix_length is False

    assert repr(part_without_end) == "500-"
    assert repr(part_without_start) == "-300"


def test_range_throws_value_error_for_invalid_value():
    with pytest.raises(ValueError):
        Range.parse("bytes=100-200, 400-600, 300-500=50-150")



import asyncio
import json
import sys
import urllib.error
import urllib.parse
import urllib.request
from asyncio import AbstractEventLoop
from typing import Any, Optional


class FailedRequestError(Exception):
    def __init__(self, status, data) -> None:
        super().__init__(
            f"The response status code does not indicate success: {status}."
            if status > -1
            else "The request failed."
        )
        self.status = status
        self.data = data


def fetch(url: str) -> Any:
    try:
        with urllib.request.urlopen(url) as response:
            return response.read()
    except urllib.error.HTTPError as http_error:
        content = http_error.read()
        raise FailedRequestError(http_error.status, _try_parse_content_as_json(content))
    except urllib.error.URLError as url_error:
        # e.g. connection refused
        raise FailedRequestError(-1, str(url_error))


def _try_parse_content_as_json(content: bytes) -> Any:
    try:
        return json.loads(content.decode("utf8"))
    except json.JSONDecodeError:
        return content


def post(url: str, data) -> Any:
    raw_data = urllib.parse.urlencode(data).encode()
    req = urllib.request.Request(
        url,
        method="POST",
        data=raw_data,
    )
    try:
        response = urllib.request.urlopen(req)
        content = response.read()
    except urllib.error.HTTPError as http_error:
        content = http_error.read()
        raise FailedRequestError(http_error.status, _try_parse_content_as_json(content))
    return _try_parse_content_as_json(content)


class HTTPHandler:
    def __init__(self, loop: Optional[AbstractEventLoop] = None) -> None:
        self._loop = loop

    @property
    def loop(self) -> AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_running_loop()
        return self._loop

    async def fetch(self, url: str) -> Any:
        return await self.loop.run_in_executor(None, lambda: fetch(url))

    async def fetch_json(self, url: str) -> Any:
        data = await self.fetch(url)
        return json.loads(data)

    async def post_form(self, url: str, data: Any) -> Any:
        return await self.loop.run_in_executor(None, lambda: post(url, data))


def get_running_loop() -> AbstractEventLoop:  # pragma: no cover
    try:
        if sys.version_info[:2] <= (3, 7):
            # For Python 3.6
            return asyncio._get_running_loop() or asyncio.get_event_loop()
        else:
            return asyncio.get_running_loop()
    except RuntimeError:
        # TODO: fix deprecation warning happening in the test suite
        # DeprecationWarning: There is no current event loop
        return asyncio.get_event_loop()



from dataclasses import dataclass
from typing import List

import pytest
from jinja2 import PackageLoader
from pydantic import BaseModel

from blacksheep.server import Application
from blacksheep.server.controllers import Controller, RoutesRegistry
from blacksheep.server.templating import (
    model_to_view_params,
    template_name,
    use_templates,
    view,
    view_async,
)
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


def get_app(enable_async):
    app = FakeApplication()
    render = use_templates(
        app,
        loader=PackageLoader("tests.testapp", "templates"),
        enable_async=enable_async,
    )
    return app, render


@pytest.fixture()
def home_model():
    return {
        "title": "Example",
        "heading": "Hello World!",
        "paragraph": "Lorem ipsum dolor sit amet",
    }


@dataclass
class Sentence:
    text: str
    url: str


@dataclass
class HelloModel:
    name: str
    sentences: List[Sentence]


class Sentence2:
    def __init__(self, text: str, url: str) -> None:
        self.text = text
        self.url = url


class HelloModel2:
    def __init__(self, name: str, sentences: List[Sentence2]) -> None:
        self.name = name
        self.sentences = sentences


class PydanticSentence(BaseModel):
    text: str
    url: str


class PydanticHelloModel(BaseModel):
    name: str
    sentences: List[PydanticSentence]


def dataclass_model():
    return HelloModel(
        "World!",
        [
            Sentence(
                "Check this out!",
                "https://github.com/Neoteroi/BlackSheep",
            )
        ],
    )


def class_model():
    return HelloModel2(
        "World!",
        [
            Sentence2(
                "Check this out!",
                "https://github.com/Neoteroi/BlackSheep",
            )
        ],
    )


def pydantic_model():
    return PydanticHelloModel(
        name="World!",
        sentences=[
            PydanticSentence(
                text="Check this out!",
                url="https://github.com/Neoteroi/BlackSheep",
            )
        ],
    )


@pytest.fixture()
def specific_text():
    return """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Specific Example</title>
</head>
<body>
    <h1>Hello World!</h1>
    <p>Lorem ipsum dolor sit amet</p>
</body>
</html>"""


nomodel_text = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Example</title>
</head>
<body>
    <h1>Hello World!</h1>
    <p>Lorem ipsum dolor sit amet</p>
</body>
</html>"""


async def _home_scenario(app: FakeApplication, url="/", expected_text=None):
    app.build_services()
    app.normalize_handlers()
    await app(get_example_scope("GET", url), MockReceive(), MockSend())
    text = await app.response.text()

    if expected_text is None:
        expected_text = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Example</title>
</head>
<body>
    <h1>Hello World!</h1>
    <p>Lorem ipsum dolor sit amet</p>
</body>
</html>"""

    assert text == expected_text
    assert app.response.status == 200


async def _view_scenario(app: FakeApplication, expected_text, url="/"):
    app.build_services()
    app.normalize_handlers()
    await app(get_example_scope("GET", url), MockReceive(), MockSend())
    text = await app.response.text()
    assert text == expected_text
    assert app.response.status == 200


@pytest.mark.asyncio
async def test_jinja_async_mode(home_model):
    app, render = get_app(True)

    @app.router.get("/")
    async def home():
        return await render("home", home_model)

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_jinja_sync_mode(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home():
        return render("home", home_model)

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_jinja_async_mode_with_verbose_method(home_model):
    app, _ = get_app(True)

    @app.router.get("/")
    async def home(jinja):
        return await view_async(jinja, "home", home_model)

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_jinja_sync_mode_with_verbose_method(home_model):
    app, _ = get_app(False)

    @app.router.get("/")
    async def home(jinja):
        return view(jinja, "home", home_model)

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_controller_conventional_view_name(home_model):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        def index(self):
            return self.view(model=home_model)

    app.setup_controllers()

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_controller_conventional_view_name_async(home_model):
    app, _ = get_app(True)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        async def index(self):
            return await self.view_async(model=home_model)

    app.setup_controllers()
    await _home_scenario(app)


@pytest.mark.asyncio
async def test_controller_specific_view_name(home_model, specific_text):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        def index(self):
            return self.view("specific", home_model)

    app.setup_controllers()

    await _home_scenario(app, expected_text=specific_text)


@pytest.mark.asyncio
async def test_controller_specific_view_name_async(home_model, specific_text):
    app, _ = get_app(True)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        async def index(self):
            return await self.view_async("specific", model=home_model)

    app.setup_controllers()
    await _home_scenario(app, expected_text=specific_text)


@pytest.mark.asyncio
async def test_controller_specific_view_name_async_no_model():
    app, _ = get_app(True)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        async def index(self):
            return await self.view_async("nomodel")

    app.setup_controllers()
    await _home_scenario(app, expected_text=nomodel_text)


@pytest.mark.asyncio
async def test_controller_conventional_view_name_no_model(home_model):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get(...)
        def nomodel(self):
            return self.view()

    app.setup_controllers()

    await _home_scenario(app, "/nomodel", expected_text=nomodel_text)


@pytest.mark.asyncio
async def test_controller_conventional_view_name_sub_function(home_model):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        def ufo(self, model):
            return self.foo(model)

        def foo(self, model):
            return self.view(model=model)

        @get()
        def index(self):
            return self.ufo(home_model)

    app.setup_controllers()

    await _home_scenario(app)


@pytest.mark.asyncio
async def test_controller_conventional_view_name_extraneous_function(home_model):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    def extraneous(controller, model):
        return controller.view(model=model)

    class Lorem(Controller):
        def ufo(self, model):
            return self.foo(model)

        def foo(self, model):
            return extraneous(self, model)

        @get()
        def index(self):
            return self.ufo(home_model)

    app.setup_controllers()

    await _home_scenario(app)


@pytest.mark.parametrize(
    "value,expected_name",
    [
        ("index", "index.html"),
        ("index.html", "index.html"),
        ("default", "default.html"),
    ],
)
def test_template_name(value, expected_name):
    assert template_name(value) == expected_name


def test_use_templates_throws_for_invalid_services():
    app = Application(services={})  # type: ignore

    with pytest.raises(TypeError):
        use_templates(
            app, loader=PackageLoader("tests.testapp", "templates"), enable_async=False
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "model_fixture",
    [
        class_model,
        dataclass_model,
        pydantic_model,
    ],
)
async def test_controller_model_interop(model_fixture):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get()
        def index(self):
            return self.view("hello", model_fixture())

    app.setup_controllers()

    await _view_scenario(
        app,
        expected_text='<div style="margin: 10em 2em;">\n  <h1>Hello, World!!</h1>\n\n'
        + '  <ul>\n    \n      <li><a href="https://github.com/Neoteroi/'
        + 'BlackSheep">Check this out!</a></li>\n    \n  </ul>\n</div>',
    )


def test_model_to_view_params_passes_unhandled_argument():
    assert model_to_view_params(2) == 2
    assert model_to_view_params("Something") == "Something"



from typing import Any, Awaitable, Callable, Optional, Sequence, Tuple

from guardpost.asynchronous.authorization import AsyncRequirement, AuthorizationStrategy
from guardpost.authorization import Policy, UnauthorizedError
from guardpost.synchronous.authorization import Requirement

from blacksheep import Request, Response, TextContent

__all__ = (
    "auth",
    "AuthorizationStrategy",
    "AuthorizationWithoutAuthenticationError",
    "allow_anonymous",
    "get_authorization_middleware",
    "Requirement",
    "AsyncRequirement",
    "handle_unauthorized",
    "Policy",
)


def auth(
    policy: Optional[str] = "authenticated",
    *,
    authentication_schemes: Optional[Sequence[str]] = None
) -> Callable[..., Any]:
    """
    Configures authorization for a decorated request handler, optionally with a policy.
    If no policy is specified, the default policy to require authenticated users is
    used.

    :param policy: optional, name of the policy to use for authorization.
    :param authentication_schemes: optional, authentication schemes to use
    for this handler. If not specified, all configured authentication handlers
    are used.
    """

    def decorator(f):
        f.auth = True
        f.auth_policy = policy
        f.auth_schemes = authentication_schemes
        return f

    return decorator


def allow_anonymous(value: bool = True) -> Callable[..., Any]:
    """
    If used without arguments, configures a decorated request handler to make it
    usable for all users: anonymous and authenticated users.

    Otherwise, enables anonymous access according to the given flag value.
    """

    def decorator(f):
        f.allow_anonymous = value
        return f

    return decorator


def get_authorization_middleware(
    strategy: AuthorizationStrategy,
) -> Callable[[Request, Callable[..., Any]], Awaitable[Response]]:
    async def authorization_middleware(request, handler):
        identity = request.identity

        if getattr(handler, "allow_anonymous", False) is True:
            return await handler(request)

        if hasattr(handler, "auth"):
            # function decorated by auth;
            await strategy.authorize(handler.auth_policy, identity)
        else:
            # function not decorated by auth; use the default policy.
            # this is necessary to support configuring authorization rules globally
            # without configuring every single request handler
            await strategy.authorize(None, identity)

        return await handler(request)

    return authorization_middleware


class AuthorizationWithoutAuthenticationError(RuntimeError):
    def __init__(self):
        super().__init__(
            "Cannot use an authorization strategy without an authentication "
            "strategy. Use `use_authentication` method to configure it."
        )


def get_www_authenticated_header_from_generic_unauthorized_error(
    error: UnauthorizedError,
) -> Optional[Tuple[bytes, bytes]]:
    if not error.scheme:
        return None

    return b"WWW-Authenticate", error.scheme.encode()


async def handle_unauthorized(
    app: Any, request: Request, http_exception: UnauthorizedError
) -> Response:
    www_authenticate = get_www_authenticated_header_from_generic_unauthorized_error(
        http_exception
    )
    return Response(
        401,
        [www_authenticate] if www_authenticate else None,
        content=TextContent("Unauthorized"),
    )



from typing import Union

from cryptography.fernet import Fernet

from . import Encryptor


class FernetEncryptor(Encryptor):
    def __init__(self, secret_key: Union[str, bytes]) -> None:
        self._fernet = Fernet(secret_key)

    def encrypt(self, value: str) -> str:
        return self._fernet.encrypt(value.encode()).decode()

    def decrypt(self, value: str) -> str:
        return self._fernet.decrypt(value.encode()).decode()



import asyncio
import os
import pathlib
import shutil
from uuid import uuid4

import pytest

from blacksheep.common.files.asyncfs import FileContext, FilesHandler


@pytest.fixture()
def files_folder():
    return pathlib.Path(__file__).parent.absolute() / "files"


@pytest.fixture(scope="module")
def temp_files_folder():
    temp_folder = pathlib.Path(__file__).parent.absolute() / ".out"
    if not temp_folder.exists():
        os.makedirs(temp_folder)
    yield temp_folder
    shutil.rmtree(temp_folder)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "file_name", ["example.txt", "pexels-photo-126407.jpeg", "README.md"]
)
async def test_read_file(files_folder: pathlib.Path, file_name: str):
    handler = FilesHandler()
    full_file_path = str(files_folder / file_name)

    contents = await handler.read(full_file_path)

    with open(full_file_path, mode="rb") as f:
        expected_contents = f.read()

    assert contents == expected_contents


@pytest.mark.asyncio
@pytest.mark.parametrize("file_name", ["example.txt", "README.md"])
async def test_read_file_rt_mode(files_folder: pathlib.Path, file_name: str):
    handler = FilesHandler()
    full_file_path = str(files_folder / file_name)

    contents = await handler.read(full_file_path, mode="rt")

    with open(full_file_path, mode="rt") as f:
        expected_contents = f.read()

    assert contents == expected_contents


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "file_name", ["example.txt", "pexels-photo-126407.jpeg", "README.md"]
)
async def test_read_file_with_open(files_folder: pathlib.Path, file_name: str):
    handler = FilesHandler()

    full_file_path = str(files_folder / file_name)
    async with handler.open(full_file_path) as file_context:
        contents = await file_context.read()

    with open(full_file_path, mode="rb") as file:
        expected_contents = file.read()

    assert contents == expected_contents


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "file_name,index,size",
    [
        ["example.txt", 0, 10],
        ["example.txt", 10, 10],
        ["example.txt", 5, 15],
        ["README.md", 0, 10],
        ["README.md", 10, 10],
        ["README.md", 5, 15],
    ],
)
async def test_seek_and_read_chunk(
    files_folder: pathlib.Path, file_name: str, index: int, size: int
):
    handler = FilesHandler()

    full_file_path = str(files_folder / file_name)
    async with handler.open(full_file_path) as file_context:
        await file_context.seek(index)
        chunk_read_async = await file_context.read(size)

    with open(full_file_path, mode="rb") as file:
        file.seek(index)
        chunk_read = file.read(size)

    assert chunk_read_async == chunk_read


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "file_name", ["example.txt", "pexels-photo-126407.jpeg", "README.md"]
)
async def test_read_file_chunks(files_folder: pathlib.Path, file_name: str):
    handler = FilesHandler()
    full_file_path = str(files_folder / file_name)

    chunk: bytes
    chunk_size = 1024
    contents = b""
    expected_contents = b""

    async with handler.open(full_file_path) as file_context:
        async for chunk in file_context.chunks(chunk_size):
            assert chunk is not None
            contents += chunk

    with open(full_file_path, mode="rb") as f:
        while True:
            chunk = f.read(chunk_size)

            if not chunk:
                break
            expected_contents += chunk

    assert contents == expected_contents


@pytest.mark.asyncio
async def test_write_file(temp_files_folder: pathlib.Path):
    handler = FilesHandler()
    file_name = str(uuid4()) + ".txt"
    full_file_path = str(temp_files_folder / file_name)

    contents = b"Lorem ipsum dolor sit"
    await handler.write(full_file_path, contents)

    with open(full_file_path, mode="rb") as f:
        expected_contents = f.read()

    assert contents == expected_contents


@pytest.mark.asyncio
async def test_write_file_text_mode(temp_files_folder: pathlib.Path):
    handler = FilesHandler()
    file_name = str(uuid4()) + ".txt"
    full_file_path = str(temp_files_folder / file_name)

    contents = "Lorem ipsum dolor sit"
    await handler.write(full_file_path, contents, mode="wt")

    with open(full_file_path, mode="rt") as f:
        expected_contents = f.read()

    assert contents == expected_contents


@pytest.mark.asyncio
async def test_write_file_with_iterable(temp_files_folder: pathlib.Path):
    handler = FilesHandler()
    file_name = str(uuid4()) + ".txt"
    full_file_path = str(temp_files_folder / file_name)

    async def provider():
        yield b"Lorem "
        await asyncio.sleep(0.01)
        yield b"ipsum"
        await asyncio.sleep(0.01)
        yield b" dolor"
        yield b" sit"

    await handler.write(full_file_path, provider)

    with open(full_file_path, mode="rb") as f:
        expected_contents = f.read()

    assert b"Lorem ipsum dolor sit" == expected_contents


@pytest.mark.asyncio
async def test_file_context_raises_for_invalid_mode():
    handler = FilesHandler()

    with pytest.raises(ValueError) as error_info:
        async with handler.open("foo.txt", mode="xx") as file_context:
            file_context.write("Foo")

    assert "invalid mode" in str(error_info.value)


@pytest.mark.asyncio
async def test_file_context_raises_if_file_is_not_open():
    with pytest.raises(TypeError) as error_info:
        file_context = FileContext("foo.txt")
        await file_context.read()

    assert str(error_info.value) == "The file is not open."



import asyncio
import ssl
from asyncio import AbstractEventLoop, TimeoutError
from typing import Any, AnyStr, Callable, Dict, List, Optional, Tuple, Type, Union, cast
from urllib.parse import urlencode

from blacksheep import URL, Content, InvalidURL, Request, Response
from blacksheep.middlewares import get_middlewares_chain
from blacksheep.utils.aio import get_running_loop

from .connection import ConnectionClosedError
from .cookies import CookieJar, cookies_middleware
from .exceptions import (
    CircularRedirectError,
    ConnectionTimeout,
    MaximumRedirectsExceededError,
    MissingLocationForRedirect,
    RequestTimeout,
    UnsupportedRedirect,
)
from .pool import ClientConnection, ClientConnectionPools

URLType = Union[str, bytes, URL]
Header = Tuple[bytes, bytes]


class RedirectsCache:
    """Used to store permanent redirects urls for later reuse"""

    __slots__ = ("_cache",)

    def __init__(self):
        self._cache: Dict[bytes, URL] = {}

    def __setitem__(self, key: bytes, value: URL):
        self._cache[key] = value

    def __getitem__(self, item: Any) -> Optional[URL]:
        try:
            return self._cache[item]
        except KeyError:
            return None

    def __contains__(self, item: Any) -> bool:
        return item in self._cache


class ClientRequestContext:
    __slots__ = ("path", "cookies")

    def __init__(self, request, cookies: Optional[CookieJar] = None):
        self.path = [request.url.value.lower()]
        self.cookies = cookies


_default_pools_by_loop_id = {}


def _get_default_pools_for_loop(loop):
    """
    This function is meant to help users of the ClientSession:
    to prevent creating many pools, decreasing performance only by
    instantiating several ClientSession without a specific HTTP Connections
    pools.
    """
    loop_id = id(loop)
    try:
        return _default_pools_by_loop_id[loop_id]
    except KeyError:
        pools = ClientConnectionPools(loop)
        _default_pools_by_loop_id[loop_id] = pools
        return pools


class ClientSession:
    def __init__(
        self,
        loop: Optional[AbstractEventLoop] = None,
        base_url: Union[None, bytes, str, URL] = None,
        ssl: Union[None, bool, ssl.SSLContext] = None,
        pools: Optional[ClientConnectionPools] = None,
        default_headers: Optional[List[Header]] = None,
        follow_redirects: bool = True,
        connection_timeout: float = 10.0,
        request_timeout: float = 60.0,
        maximum_redirects: int = 20,
        redirects_cache_type: Union[Type[RedirectsCache], Any] = None,
        cookie_jar: Union[None, bool, CookieJar] = None,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        if loop is None:
            loop = get_running_loop()

        if not pools:
            # TODO: maybe it is safer to emit a warning here,
            # and create an instance of ClientConnectionPools
            pools = _get_default_pools_for_loop(loop)
            self.owns_pools = False
        else:
            self.owns_pools = True

        if redirects_cache_type is None and follow_redirects:
            redirects_cache_type = RedirectsCache

        if middlewares is None:
            middlewares = []

        if cookie_jar is None:
            cookie_jar = CookieJar()

        if cookie_jar is False:
            cookie_jar = None
        else:
            middlewares.insert(0, cookies_middleware)

        self.loop = loop
        self._base_url: Optional[URL]
        self.base_url = base_url
        self.ssl = ssl
        self.default_headers = default_headers
        self.pools = pools
        self.connection_timeout = connection_timeout
        self.request_timeout = request_timeout
        self.follow_redirects = follow_redirects
        self.cookie_jar: Optional[CookieJar] = cast(Optional[CookieJar], cookie_jar)
        self._permanent_redirects_urls = (
            redirects_cache_type() if follow_redirects else None
        )
        self.non_standard_handling_of_301_302_redirect_method = True
        self.maximum_redirects = maximum_redirects
        self._handler = None
        self._middlewares: List[Callable[..., Any]]
        self.middlewares = middlewares
        self.delay_before_retry = 0.5

    @property
    def middlewares(self) -> List[Callable[..., Any]]:
        return self._middlewares

    @middlewares.setter
    def middlewares(self, value) -> None:
        self._middlewares = list(value)
        self._build_middlewares_chain()

    @property
    def base_url(self) -> Optional[URL]:
        return self._base_url

    @base_url.setter
    def base_url(self, value):
        url = None
        if value and not isinstance(value, URL):
            if isinstance(value, str):
                url = URL(value.encode())
            else:
                url = URL(value)
        self._base_url = url

    def add_middlewares(self, middlewares: List[Callable]):
        self._middlewares += middlewares
        self._build_middlewares_chain()

    def _build_middlewares_chain(self):
        if not self.middlewares:
            return

        async def root_handler(request):
            return await self._send_core(request)

        self._handler = get_middlewares_chain(self._middlewares, root_handler)

    def use_standard_redirect(self):
        """Uses specification compliant handling of 301 and 302 redirects"""
        self.non_standard_handling_of_301_302_redirect_method = False

    def get_url(self, url, params=None) -> bytes:
        value = self.get_url_value(url)
        if not params:
            return value
        query = urlencode(params).encode("ascii")
        return value + (b"&" if b"?" in value else b"?") + query

    def get_url_value(self, url: Union[AnyStr, URL]) -> bytes:
        if isinstance(url, str):
            url = url.encode()

        if not isinstance(url, URL):
            if url == b"":
                url = b"/"
            url = URL(url)

        if url.is_absolute:
            return url.value

        if self.base_url:
            return self.base_url.join(url).value
        return url.value

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        if self.owns_pools:
            self.pools.dispose()

    @staticmethod
    def extract_redirect_location(response: Response) -> URL:
        # if the server returned more than one value, use
        # the first header in order
        location = response.get_first_header(b"Location")
        if not location:
            raise MissingLocationForRedirect(response)

        # if the location cannot be parsed as URL, let exception happen:
        # this might be a redirect to a URN!
        # simply don't follows the redirect, and returns the response to
        # the caller
        try:
            return URL(location)
        except InvalidURL:
            raise UnsupportedRedirect(location)

    @staticmethod
    def get_redirect_url(request: Request, location: URL) -> URL:
        if location.is_absolute:
            return location
        # relative redirect URI
        # https://tools.ietf.org/html/rfc7231#section-7.1.2
        return request.url.base_url().join(location)

    def validate_redirect(
        self, redirect_url: URL, response: Response, context: ClientRequestContext
    ):
        redirect_url_lower = redirect_url.value.lower()
        if redirect_url_lower in context.path:
            context.path.append(redirect_url_lower)
            raise CircularRedirectError(context.path, response)

        context.path.append(redirect_url_lower)

        if len(context.path) > self.maximum_redirects:
            raise MaximumRedirectsExceededError(
                context.path, response, self.maximum_redirects
            )

    def update_request_for_redirect(self, request: Request, response: Response) -> None:
        context: ClientRequestContext = getattr(
            request, "context", ClientRequestContext(request, self.cookie_jar)
        )
        status = response.status

        if status == 301 or status == 302:
            if self.non_standard_handling_of_301_302_redirect_method:
                # Change original request method to GET (Browser-like)
                request.method = "GET"

        if status == 303:
            # 303 See Other
            # Change original request method to GET
            request.method = "GET"

        location = self.extract_redirect_location(response)
        redirect_url = self.get_redirect_url(request, location)

        self.validate_redirect(redirect_url, response, context)

        if status == 301 or status == 308:
            self._permanent_redirects_urls[request.url.value] = redirect_url

        request.url = redirect_url

    def merge_default_headers(self, request: Request) -> None:
        if not self.default_headers:
            return

        for header in self.default_headers:
            # TODO: support tuple and Header?
            if header[0] not in request.headers:
                request.headers.add(header[0], header[1])

    def check_permanent_redirects(self, request: Request) -> None:
        if (
            self._permanent_redirects_urls is not None
            and request.url.value in self._permanent_redirects_urls
        ):
            redirect_url = self._permanent_redirects_urls[request.url.value]
            if redirect_url is not None:
                request.url = redirect_url

    async def get_connection(self, url: URL) -> ClientConnection:
        pool = self.pools.get_pool(url.schema, url.host, url.port, self.ssl)

        try:
            return await asyncio.wait_for(
                pool.get_connection(), self.connection_timeout
            )
        except TimeoutError:
            raise ConnectionTimeout(url.base_url(), self.connection_timeout)

    def get_new_context(self, request: Request) -> ClientRequestContext:
        return ClientRequestContext(request, self.cookie_jar)

    def _validate_request_url(self, request: Request):
        if not request.url.is_absolute:
            if self.base_url:
                request.url = URL(self.get_url_value(request.url))
            else:
                raise ValueError(
                    "request.url must be a complete, absolute URL. "
                    "Either provide a base_url "
                    "for the client, or specify a full URL for the request."
                )

    async def send(self, request: Request) -> Response:
        self._validate_request_url(request)

        if not hasattr(request, "context"):
            request.context = self.get_new_context(request)  # type: ignore
            self.merge_default_headers(request)

        if self._handler:
            # using middlewares
            return await self._handler(request)

        # without middlewares
        return await self._send_core(request)

    async def _send_core(self, request: Request) -> Response:
        self.check_permanent_redirects(request)

        response = await self._send_using_connection(request)

        if self.follow_redirects and response.is_redirect():
            try:
                self.update_request_for_redirect(request, response)
            except UnsupportedRedirect:
                # redirect not to HTTP / HTTPS: for example,
                # it can be a redirect to a URN
                return response
            return await self.send(request)

        return response

    async def _send_using_connection(self, request, attempt: int = 1) -> Response:
        connection = await self.get_connection(request.url)

        try:
            return await asyncio.wait_for(
                connection.send(request), self.request_timeout
            )
        except ConnectionClosedError as connection_closed_error:
            if connection_closed_error.can_retry and attempt < 4:
                await asyncio.sleep(self.delay_before_retry)
                return await self._send_using_connection(request, attempt + 1)
            raise
        except TimeoutError:
            raise RequestTimeout(request.url, self.request_timeout)

    async def get(
        self, url: URLType, headers: Optional[List[Header]] = None, params=None
    ) -> Response:
        return await self.send(Request("GET", self.get_url(url, params), headers))

    async def post(
        self,
        url: URLType,
        content: Content = None,
        headers: Optional[List[Header]] = None,
        params=None,
    ) -> Response:
        request = Request("POST", self.get_url(url, params), headers)
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def put(
        self,
        url: URLType,
        content: Content = None,
        headers: Optional[List[Header]] = None,
        params=None,
    ) -> Response:
        request = Request("PUT", self.get_url(url, params), headers)
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def delete(
        self,
        url: URLType,
        content: Content = None,
        headers: Optional[List[Header]] = None,
        params=None,
    ) -> Response:
        request = Request("DELETE", self.get_url(url, params), headers)
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def trace(
        self, url: URLType, headers: Optional[List[Header]] = None, params=None
    ) -> Response:
        return await self.send(Request("TRACE", self.get_url(url, params), headers))

    async def head(
        self, url: URLType, headers: Optional[List[Header]] = None, params=None
    ) -> Response:
        return await self.send(Request("HEAD", self.get_url(url, params), headers))

    async def patch(
        self,
        url: URLType,
        content: Content = None,
        headers: Optional[List[Header]] = None,
        params=None,
    ) -> Response:
        request = Request("PATCH", self.get_url(url, params), headers)
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def options(
        self,
        url: URLType,
        content: Content = None,
        headers: Optional[List[Header]] = None,
        params=None,
    ) -> Response:
        request = Request("OPTIONS", self.get_url(url, params), headers)
        return await self.send(
            request.with_content(content) if content is not None else request
        )



from datetime import datetime
from typing import Any, Optional, Sequence

from guardpost.asynchronous.authentication import AuthenticationHandler
from guardpost.authentication import Identity
from itsdangerous import Serializer
from itsdangerous.exc import BadSignature

from blacksheep.baseapp import get_logger
from blacksheep.cookies import Cookie
from blacksheep.messages import Request, Response
from blacksheep.server.dataprotection import get_serializer
from blacksheep.utils import ensure_str


class CookieAuthentication(AuthenticationHandler):
    """
    An AuthenticationHandler that tries to restore the user's context from a cookie.
    """

    def __init__(
        self,
        cookie_name: str = "identity",
        secret_keys: Optional[Sequence[str]] = None,
        serializer: Optional[Serializer] = None,
        auth_scheme: Optional[str] = None,
    ) -> None:
        """
        Creates a new instance of CookieAuthentication handler, that tries to obtain
        the user's identity by cookie.

        Parameters
        ----------
        cookie_name : str, optional
            The name of the cookie used to restore user's identity, by default
            "identity"
        secret_key : str, optional
            If specified, the key used by a default serializer (when no serializer is
            specified), by default None
        serializer : Optional[Serializer], optional
            If specified, controls the serializer used to sign and verify the values
            of cookies used for identities, by default None
        auth_scheme : str, optional
            The name of the authentication scheme declared for users' identity, by
            default f"CookieAuth: {cookie_name}"
        """
        super().__init__()
        self.cookie_name = cookie_name
        self.serializer = serializer or get_serializer(
            secret_keys, f"{cookie_name}auth"
        )
        self.auth_scheme = auth_scheme or f"CookieAuth: {cookie_name}"
        self.logger = get_logger()

    def set_cookie(self, data: Any, response: Response, secure: bool = False) -> None:
        """
        Sets the cookie used for authentication. If a cookie is set, it is assumed that
        the user is recognized (authenticated). The passed value is serialized, meaning
        signed and encrypted using the `itsdangerous.Serializer` associated with this
        CookieAuthentication handler. A common scenario is that data is a dictionary
        with claims describing the identity of the user (e.g. id_token claims).

        Parameters
        ----------
        data : Any
            Anything that can be serialized by an `itsdangerous.Serializer`, a
            dictionary in the most common scenario.
        response : Response
            The instance of blacksheep `Response` that will include the cookie for the
            client.
        secure : bool, optional
            Whether the set cookie should have secure flag, by default False
        """
        value = self.serializer.dumps(data)

        response.set_cookie(
            Cookie(
                self.cookie_name,
                ensure_str(value),  # type: ignore
                domain=None,
                path="/",
                http_only=True,
                secure=secure,
                expires=datetime.fromtimestamp(data["exp"]) if "exp" in data else None,
            )
        )

    def unset_cookie(self, response: Response) -> None:
        """
        Unsets the cookie used for authentication.
        """
        response.unset_cookie(self.cookie_name)

    def set_user_context(self, context: Request, data: Any) -> None:
        """
        Sets the user context, when user's data was parsed and validated from a cookie.
        """
        context.identity = Identity(data, self.auth_scheme)

    async def authenticate(self, context: Request) -> Optional[Identity]:
        cookie = context.get_cookie(self.cookie_name)

        if cookie is None:
            context.identity = Identity({})
        else:
            try:
                value = self.serializer.loads(cookie)
            except BadSignature:
                self.logger.debug(
                    "Cookie authentication failed (%s), invalid signature.",
                    self.cookie_name,
                )
                context.identity = Identity({})
            else:
                self.set_user_context(context, value)
        return None



from typing import Optional

from blacksheep.contents import Content
from blacksheep.server.application import Application
from blacksheep.server.responses import Response
from blacksheep.testing.simulator import AbstractTestSimulator, TestSimulator

from .helpers import CookiesType, HeadersType, QueryType


class TestClient:
    # Setting this dunder variable
    # We tell to pytest don't discover this up
    __test__ = False

    def __init__(
        self, app: Application, test_simulator: Optional[AbstractTestSimulator] = None
    ):
        self._test_simulator = test_simulator or TestSimulator(app)

    async def get(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP GET method"""
        return await self._test_simulator.send_request(
            method="GET",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=None,
        )

    async def post(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP POST method"""
        return await self._test_simulator.send_request(
            method="POST",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=content,
        )

    async def patch(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP PATCH method"""
        return await self._test_simulator.send_request(
            method="PATCH",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=content,
        )

    async def put(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP PUT method"""
        return await self._test_simulator.send_request(
            method="PUT",
            path=path,
            headers=headers,
            query=query,
            content=content,
            cookies=cookies,
        )

    async def delete(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP DELETE method"""
        return await self._test_simulator.send_request(
            method="DELETE",
            path=path,
            headers=headers,
            query=query,
            content=content,
            cookies=cookies,
        )

    async def options(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP OPTIONS method"""
        return await self._test_simulator.send_request(
            method="OPTIONS",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    async def head(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP HEAD method"""
        return await self._test_simulator.send_request(
            method="HEAD",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    async def trace(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP TRACE method"""
        return await self._test_simulator.send_request(
            method="TRACE",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )



from typing import List

import pytest

from blacksheep.server.openapi.common import ParameterInfo
from blacksheep.server.openapi.docstrings import (
    DocstringInfo,
    EpytextDialect,
    GoogleDocDialect,
    NumpydocDialect,
    ReStructuredTextDialect,
    collapse,
)


@pytest.mark.parametrize(
    "docstring,expected_info",
    [
        (
            """
            Example with lists of values.

            @param List[str] a: list of str
            @param int[] b: list of int
            @param str[] c: list of str
            """,
            DocstringInfo(
                summary="Example with lists of values.",
                description="Example with lists of values.",
                parameters={
                    "a": ParameterInfo("list of str", value_type=List[str]),
                    "b": ParameterInfo("list of int", value_type=List[int]),
                    "c": ParameterInfo("list of str", value_type=List[str]),
                },
            ),
        ),
        (
            """
            Example with lists of values.

            @param a: list of str
            @param b: list of int
            @param c: list of str
            @type a: List[str]
            @type b: int[]
            @type c: str[]
            """,
            DocstringInfo(
                summary="Example with lists of values.",
                description="Example with lists of values.",
                parameters={
                    "a": ParameterInfo("list of str", value_type=List[str]),
                    "b": ParameterInfo("list of int", value_type=List[int]),
                    "c": ParameterInfo("list of str", value_type=List[str]),
                },
            ),
        ),
        (
            """
            Return the x intercept of the line M{y=m*x+b}. The X{x intercept}
            of a line is the point at which it crosses the x axis (M{y=0}).

            This function can be used in conjuction with L{z_transform} to
            find an arbitrary function's zeros.

            @param number m: The slope of the line.
            @param number b: The y intercept of the line.  The X{y intercept} of a
                    line is the point at which it crosses the y axis (M{x=0}).
            @rtype:   number
            @return:  the x intercept of the line M{y=m*x+b}.
            """,
            DocstringInfo(
                summary="Return the x intercept of the line M{y=m*x+b}. "
                + "The X{x intercept} of a line is the point at which it "
                + "crosses the x axis (M{y=0}).",
                description="Return the x intercept of the line M{y=m*x+b}. "
                + "The X{x intercept} of a line is the point at which it "
                + "crosses the x axis (M{y=0}). This function can be used in "
                + "conjuction with L{z_transform} to find an arbitrary function's zeros.",
                parameters={
                    "m": ParameterInfo("The slope of the line.", value_type=float),
                    "b": ParameterInfo(
                        "The y intercept of the line. The X{y intercept} "
                        + "of a line is the point at which it crosses the y axis (M{x=0}).",
                        value_type=float,
                    ),
                },
                return_type=float,
                return_description="the x intercept of the line M{y=m*x+b}.",
            ),
        ),
        (
            """
            Return the x intercept of the line M{y=m*x+b}. The X{x intercept}
            of a line is the point at which it crosses the x axis (M{y=0}).

            This function can be used in conjuction with L{z_transform} to
            find an arbitrary function's zeros.

            @type  m: number
            @param m: The slope of the line.
            @type  b: number
            @param b: The y intercept of the line.  The X{y intercept} of a
                    line is the point at which it crosses the y axis (M{x=0}).
            @rtype:   number
            @return:  the x intercept of the line M{y=m*x+b}.
            """,
            DocstringInfo(
                summary="Return the x intercept of the line M{y=m*x+b}. "
                + "The X{x intercept} of a line is the point at which it "
                + "crosses the x axis (M{y=0}).",
                description="Return the x intercept of the line M{y=m*x+b}. "
                + "The X{x intercept} of a line is the point at which it "
                + "crosses the x axis (M{y=0}). This function can be used in "
                + "conjuction with L{z_transform} to find an arbitrary function's zeros.",
                parameters={
                    "m": ParameterInfo("The slope of the line.", value_type=float),
                    "b": ParameterInfo(
                        "The y intercept of the line. The X{y intercept} "
                        + "of a line is the point at which it crosses the y axis (M{x=0}).",
                        value_type=float,
                    ),
                },
                return_type=float,
                return_description="the x intercept of the line M{y=m*x+b}.",
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet.

            @type  value: int
            @param value: Some value.
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet.",
                description="Lorem ipsum dolor sit amet.",
                parameters={
                    "value": ParameterInfo("Some value.", value_type=int),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet.

            @type  value: int or None
            @param value: Some value.
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet.",
                description="Lorem ipsum dolor sit amet.",
                parameters={
                    "value": ParameterInfo(
                        "Some value.", value_type=int, required=False
                    ),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet.

            @type  value: int?
            @param value: Some value.
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet.",
                description="Lorem ipsum dolor sit amet.",
                parameters={
                    "value": ParameterInfo(
                        "Some value.", value_type=int, required=False
                    ),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            This is a javadoc style.

            @param param1: this is a first param
            @param param2: this is a second param
            @return: this is a description of what is returned
            @raise keyError: raises an exception
            """,
            DocstringInfo(
                summary="This is a javadoc style.",
                description="This is a javadoc style.",
                parameters={
                    "param1": ParameterInfo("this is a first param"),
                    "param2": ParameterInfo("this is a second param"),
                },
                return_description="this is a description of what is returned",
            ),
        ),
        (
            """
            This is a paragraph. Paragraphs can
            span multiple lines, and can contain
            I{inline markup}.

            This is another paragraph.  Paragraphs
            are separated by blank lines.
            """,
            DocstringInfo(
                summary=collapse(
                    """
                    This is a paragraph. Paragraphs can
                    span multiple lines, and can contain
                    I{inline markup}.
                    """
                ),
                description=collapse(
                    """
                    This is a paragraph. Paragraphs can
                    span multiple lines, and can contain
                    I{inline markup}.

                    This is another paragraph.  Paragraphs
                    are separated by blank lines.
                    """
                ),
                parameters={},
            ),
        ),
    ],
)
def test_epytext_dialect(docstring, expected_info):
    dialect = EpytextDialect()

    info = dialect.parse_docstring(docstring)
    assert expected_info == info


@pytest.mark.parametrize(
    "docstring,expected_info",
    [
        (
            """
            Example with lists of values.

            :param List[str] a: list of str
            :param int[] b: list of int
            :param str[] c: list of str
            """,
            DocstringInfo(
                summary="Example with lists of values.",
                description="Example with lists of values.",
                parameters={
                    "a": ParameterInfo("list of str", value_type=List[str]),
                    "b": ParameterInfo("list of int", value_type=List[int]),
                    "c": ParameterInfo("list of str", value_type=List[str]),
                },
            ),
        ),
        (
            """
            Example with unrecognized type.

            :param foo a: some unrecognized type
            """,
            DocstringInfo(
                summary="Example with unrecognized type.",
                description="Example with unrecognized type.",
                parameters={
                    "a": ParameterInfo("some unrecognized type (foo)", value_type=None),
                },
            ),
        ),
        (
            """
            Example with lists of values.

            :param a: list of str
            :param b: list of int
            :param c: list of str
            :type a: List[str]
            :type b: int[]
            :type c: str[]
            """,
            DocstringInfo(
                summary="Example with lists of values.",
                description="Example with lists of values.",
                parameters={
                    "a": ParameterInfo("list of str", value_type=List[str]),
                    "b": ParameterInfo("list of int", value_type=List[int]),
                    "c": ParameterInfo("list of str", value_type=List[str]),
                },
            ),
        ),
        (
            """
            Send a message to a recipient

            :param sender: The person sending the message
            :param recipient: The recipient of the message
            :param message_body: The body of the message
            :param priority: The priority of the message, can be a number 1-5
            :type priority: integer or None
            :return: the message id
            :rtype: int
            :raises ValueError: if the message_body exceeds 160 characters
            :raises TypeError: if the message_body is not a basestring
            """,
            DocstringInfo(
                summary="Send a message to a recipient",
                description="Send a message to a recipient",
                parameters={
                    "sender": ParameterInfo(
                        "The person sending the message", value_type=None
                    ),
                    "recipient": ParameterInfo(
                        "The recipient of the message", value_type=None
                    ),
                    "message_body": ParameterInfo(
                        "The body of the message", value_type=None
                    ),
                    "priority": ParameterInfo(
                        "The priority of the message, can be a number 1-5",
                        value_type=int,
                        required=False,
                    ),
                },
                return_type=int,
                return_description="the message id",
            ),
        ),
        (
            """
            Send a message to a recipient

            :param str sender: The person sending the message
            :param str recipient: The recipient of the message
            :param str message_body: The body of the message
            :param integer or None priority: The priority of the message, can be a number 1-5
            :return: the message id
            :rtype: int
            :raises ValueError: if the message_body exceeds 160 characters
            :raises TypeError: if the message_body is not a basestring
            """,
            DocstringInfo(
                summary="Send a message to a recipient",
                description="Send a message to a recipient",
                parameters={
                    "sender": ParameterInfo(
                        "The person sending the message", value_type=str
                    ),
                    "recipient": ParameterInfo(
                        "The recipient of the message", value_type=str
                    ),
                    "message_body": ParameterInfo(
                        "The body of the message", value_type=str
                    ),
                    "priority": ParameterInfo(
                        "The priority of the message, can be a number 1-5",
                        value_type=int,
                        required=False,
                    ),
                },
                return_type=int,
                return_description="the message id",
            ),
        ),
    ],
)
def test_rest_dialect(docstring, expected_info):
    dialect = ReStructuredTextDialect()

    info = dialect.parse_docstring(docstring)
    assert expected_info == info


@pytest.mark.parametrize(
    "docstring,expected_info",
    [
        (
            """
            My numpydoc description of a kind
            of very exhautive numpydoc format docstring.

            Parameters
            ----------
            first : str
                the 1st param name `first`
            second :
                the 2nd param
            third : int, optional
                the 3rd param, by default 'value'

            Returns
            -------
            string
                a value in a string

            Raises
            ------
            KeyError
                when a key error
            OtherError
                when an other error
            """,
            DocstringInfo(
                summary="My numpydoc description of a kind of very exhautive "
                + "numpydoc format docstring.",
                description="My numpydoc description of a kind of very exhautive "
                + "numpydoc format docstring.",
                parameters={
                    "first": ParameterInfo(
                        "the 1st param name `first`", value_type=str
                    ),
                    "second": ParameterInfo("the 2nd param", value_type=None),
                    "third": ParameterInfo(
                        "the 3rd param, by default 'value'",
                        value_type=int,
                        required=False,
                    ),
                },
                return_type=str,
                return_description="a value in a string",
            ),
        ),
        (
            """
            Example with lists of values.

            Parameters
            ----------
            a : List[str]
                list of str
            b : int[]
                list of int
            c : str[]
                list of str
            """,
            DocstringInfo(
                summary="Example with lists of values.",
                description="Example with lists of values.",
                parameters={
                    "a": ParameterInfo("list of str", value_type=List[str]),
                    "b": ParameterInfo("list of int", value_type=List[int]),
                    "c": ParameterInfo("list of str", value_type=List[str]),
                },
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
            tempor incididunt ut labore et dolore magna aliqua.

            Ut enim ad minim  veniam, quis nostrud exercitation ullamco laboris nisi ut
            aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in
            voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint
            occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit
            anim id est laborum.

            Parameters
            ----------
            lorem : str
                A very long description spanning across multiple lines;
                Ut enim ad minim  veniam, quis nostrud exercitation ullamco laboris nisi
                ut aliquip ex ea commodo consequat. Duis aute irure dolor in
                reprehenderit.
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
                + "do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
                description="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
                + "do eiusmod tempor incididunt ut labore et dolore magna aliqua. "
                + "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris "
                + "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in "
                + "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla "
                + "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in "
                + "culpa qui officia deserunt mollit anim id est laborum.",
                parameters={
                    "lorem": ParameterInfo(
                        collapse(
                            """
                            A very long description spanning across multiple lines;
                            Ut enim ad minim veniam, quis nostrud exercitation ullamco
                            laboris nisi ut aliquip ex ea commodo consequat. Duis aute
                            irure dolor in reprehenderit."""
                        ),
                        value_type=str,
                    ),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            My numpydoc description of a kind
            of very exhautive numpydoc format docstring.

            Params
            ----------
            first : str
                the 1st param name `first`
            second :
                the 2nd param
            third : int, optional
                the 3rd param, by default 'value'

            Returns
            -------
            string
                a value in a string

            Raises
            ------
            KeyError
                when a key error
            OtherError
                when an other error
            """,
            DocstringInfo(
                summary="My numpydoc description of a kind of very exhautive "
                + "numpydoc format docstring.",
                description="My numpydoc description of a kind of very exhautive "
                + "numpydoc format docstring.",
                parameters={
                    "first": ParameterInfo(
                        "the 1st param name `first`", value_type=str
                    ),
                    "second": ParameterInfo("the 2nd param", value_type=None),
                    "third": ParameterInfo(
                        "the 3rd param, by default 'value'",
                        value_type=int,
                        required=False,
                    ),
                },
                return_type=str,
                return_description="a value in a string",
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet 2.

            Parameters
            ----------
            filename : str
            copy : bool
            dtype : data-type
            iterable : iterable object
            shape : int or tuple of int
            files : list of str

            Returns
            -------
            string
                a value in a string

            Raises
            ------
            KeyError
                when a key error
            OtherError
                when an other error
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet 2.",
                description="Lorem ipsum dolor sit amet 2.",
                parameters={
                    "filename": ParameterInfo("", value_type=str),
                    "copy": ParameterInfo("", value_type=bool),
                    "dtype": ParameterInfo("data-type", value_type=None),
                    "iterable": ParameterInfo("iterable object", value_type=None),
                    "shape": ParameterInfo("int or tuple of int", value_type=None),
                    "files": ParameterInfo("list of str", value_type=None),
                },
                return_type=str,
                return_description="a value in a string",
            ),
        ),
        (
            """
            Lorem ipsum dolor sit amet.

            Custom
            ----------
            Hello World!
            """,
            DocstringInfo(
                summary="Lorem ipsum dolor sit amet.",
                description="Lorem ipsum dolor sit amet.",
                parameters={},
            ),
        ),
    ],
)
def test_numpydoc_dialect(docstring, expected_info):
    dialect = NumpydocDialect()

    info = dialect.parse_docstring(docstring)
    assert expected_info == info


def test_numpydoc_dialect_warns_about_invalid_parameter():
    docstring = """
    Lorem ipsum dolor sit amet.

    Parameters
    ----------
    something_invalid
    """

    with pytest.warns(
        UserWarning,
        match="Invalid parameter definition in docstring: something_invalid",
    ):
        dialect = NumpydocDialect()
        info = dialect.parse_docstring(docstring)
        assert info is not None


def test_googledoc_dialect_warns_about_invalid_parameter():
    docstring = """
    Lorem ipsum dolor sit amet.

    Args:
        something_invalid
    """

    with pytest.warns(
        UserWarning,
        match="Invalid parameter definition in docstring: something_invalid",
    ):
        dialect = GoogleDocDialect()
        info = dialect.parse_docstring(docstring)
        assert info is not None


@pytest.mark.parametrize(
    "docstring,expected_info",
    [
        (
            """
            This is an example of Google style.

            Args:
                param1: This is the first param.
                param2: This is a second param.

            Returns:
                This is a description of what is returned.

            Raises:
                KeyError: Raises an exception.
            """,
            DocstringInfo(
                summary="This is an example of Google style.",
                description="This is an example of Google style.",
                parameters={
                    "param1": ParameterInfo("This is the first param."),
                    "param2": ParameterInfo("This is a second param."),
                },
                return_type=None,
                return_description="This is a description of what is returned.",
            ),
        ),
        (
            """
            This is an example of Google style.

            Args:
                param1: This is the first param.
                param2: This is a second param.

            Raises:
                KeyError: Raises an exception.
            """,
            DocstringInfo(
                summary="This is an example of Google style.",
                description="This is an example of Google style.",
                parameters={
                    "param1": ParameterInfo("This is the first param."),
                    "param2": ParameterInfo("This is a second param."),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            This is an example of Google style.

            Args:
                param1: This is the first param.
                param2: This is a second param.
            """,
            DocstringInfo(
                summary="This is an example of Google style.",
                description="This is an example of Google style.",
                parameters={
                    "param1": ParameterInfo("This is the first param."),
                    "param2": ParameterInfo("This is a second param."),
                },
                return_type=None,
                return_description=None,
            ),
        ),
        (
            """
            Fetches rows from a Smalltable.

            Retrieves rows pertaining to the given keys from the Table instance
            represented by table_handle. String keys will be UTF-8 encoded.

            Args:
                table_handle:
                    An open smalltable.Table instance.
                keys:
                    A sequence of strings representing the key of each table row to
                    fetch. String keys will be UTF-8 encoded.
                require_all_keys:
                    Optional; If require_all_keys is True only rows with values set
                    for all keys will be returned.

            Returns:
                A dict mapping keys to the corresponding table row data
                fetched. Each row is represented as a tuple of strings. For
                example:

                {b'Serak': ('Rigel VII', 'Preparer'),
                b'Zim': ('Irk', 'Invader'),
                b'Lrrr': ('Omicron Persei 8', 'Emperor')}

                Returned keys are always bytes. If a key from the keys argument is
                missing from the dictionary, then that row was not found in the
                table (and require_all_keys must have been False).

            Raises:
                IOError: An error occurred accessing the smalltable.
            """,
            DocstringInfo(
                summary="Fetches rows from a Smalltable.",
                description="Fetches rows from a Smalltable. Retrieves rows "
                + "pertaining to the given keys from the Table instance represented "
                + "by table_handle. String keys will be UTF-8 encoded.",
                parameters={
                    "table_handle": ParameterInfo("An open smalltable.Table instance."),
                    "keys": ParameterInfo(
                        "A sequence of strings representing the key of each table "
                        "row to fetch. String keys will be UTF-8 encoded."
                    ),
                    "require_all_keys": ParameterInfo(
                        "Optional; If require_all_keys is True only rows with values set for "
                        "all keys will be returned.",
                    ),
                },
                return_type=None,
                return_description=collapse(
                    """
                    A dict mapping keys to the corresponding table row data
                    fetched. Each row is represented as a tuple of strings. For
                    example:

                    {b'Serak': ('Rigel VII', 'Preparer'),
                    b'Zim': ('Irk', 'Invader'),
                    b'Lrrr': ('Omicron Persei 8', 'Emperor')}

                    Returned keys are always bytes. If a key from the keys argument is
                    missing from the dictionary, then that row was not found in the
                    table (and require_all_keys must have been False).
                    """
                ),
            ),
        ),
    ],
)
def test_googledoc_dialect(docstring, expected_info):
    dialect = GoogleDocDialect()

    info = dialect.parse_docstring(docstring)
    assert expected_info == info



import pytest

from blacksheep.exceptions import BadRequest
from blacksheep.server.application import ApplicationAlreadyStartedCORSError
from blacksheep.server.cors import (
    CORSConfigurationError,
    CORSPolicy,
    CORSPolicyNotConfiguredError,
    CORSStrategy,
    NotRequestHandlerError,
)
from blacksheep.server.responses import text
from blacksheep.server.routing import Router
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend


def test_app_raises_type_error_for_cors_property(app):
    with pytest.raises(TypeError):
        app.cors


def test_cors_policy():
    policy = CORSPolicy(
        allow_methods=["GET", "POST", "DELETE"],
        allow_headers=["Authorization"],
        allow_origins=["http://localhost:44555"],
    )
    assert policy.allow_methods == {"GET", "POST", "DELETE"}
    assert policy.allow_headers == {"authorization"}
    assert policy.allow_origins == {"http://localhost:44555"}


def test_cors_policy_setters_strings():
    policy = CORSPolicy()

    policy.allow_methods = "get delete"
    assert policy.allow_methods == {"GET", "DELETE"}

    policy.allow_methods = "GET POST PATCH"
    assert policy.allow_methods == {"GET", "POST", "PATCH"}

    policy.allow_methods = "GET, POST, PATCH"
    assert policy.allow_methods == {"GET", "POST", "PATCH"}

    policy.allow_methods = "GET,POST,PATCH"
    assert policy.allow_methods == {"GET", "POST", "PATCH"}

    policy.allow_methods = "GET;POST;PATCH"
    assert policy.allow_methods == {"GET", "POST", "PATCH"}

    for value in {"X-Foo Authorization", "X-Foo, Authorization", "X-Foo,Authorization"}:
        policy.allow_headers = value
        assert policy.allow_headers == {"x-foo", "authorization"}

    policy.allow_origins = "http://Example.com https://Bezkitu.ORG"
    assert policy.allow_origins == {"http://example.com", "https://bezkitu.org"}

    policy.allow_headers = None
    assert policy.allow_headers == frozenset()

    policy.allow_methods = None
    assert policy.allow_methods == frozenset()

    policy.allow_origins = None
    assert policy.allow_origins == frozenset()


def test_cors_policy_setters_force_case():
    policy = CORSPolicy()

    policy.allow_methods = ["get", "delete"]
    assert policy.allow_methods == {"GET", "DELETE"}

    policy.allow_headers = ["X-Foo", "Authorization"]
    assert policy.allow_headers == {"x-foo", "authorization"}

    policy.allow_origins = ["http://Example.com", "https://Bezkitu.ORG"]
    assert policy.allow_origins == {"http://example.com", "https://bezkitu.org"}


def test_cors_policy_allow_all_methods():
    policy = CORSPolicy()

    assert policy.allow_headers == set()
    policy.allow_any_header()
    assert policy.allow_headers == {"*"}

    assert policy.allow_methods == set()
    policy.allow_any_method()
    assert policy.allow_methods == {"*"}

    assert policy.allow_origins == set()
    policy.allow_any_origin()
    assert policy.allow_origins == {"*"}


def test_cors_policy_raises_for_negative_max_age():
    with pytest.raises(ValueError):
        CORSPolicy(max_age=-1)

    policy = CORSPolicy()
    with pytest.raises(ValueError):
        policy.max_age = -5


def test_cors_strategy_raises_for_missing_policy_name():
    cors = CORSStrategy(CORSPolicy(), Router())

    with pytest.raises(CORSConfigurationError):
        cors.add_policy("", CORSPolicy())

    with pytest.raises(CORSConfigurationError):
        cors.add_policy(None, CORSPolicy())  # type: ignore


def test_cors_strategy_raises_for_duplicate_policy_name():
    cors = CORSStrategy(CORSPolicy(), Router())

    cors.add_policy("a", CORSPolicy())

    with pytest.raises(CORSConfigurationError):
        cors.add_policy("a", CORSPolicy())


@pytest.mark.asyncio
async def test_cors_request(app):
    app.use_cors(
        allow_methods="GET POST DELETE", allow_origins="https://www.neoteroi.dev"
    )

    @app.router.get("/")
    async def home():
        return text("Hello, World")

    @app.router.put("/")
    async def put_something():
        ...

    await app.start()

    await app(
        get_example_scope("GET", "/", [(b"Origin", b"https://www.something-else.dev")]),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        response.headers.get_single(b"CORS-Error")
        == b"The origin of the request is not enabled by CORS rules."
    )

    await app(
        get_example_scope("PUT", "/", [(b"Origin", b"https://www.neoteroi.dev")]),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        response.headers.get_single(b"CORS-Error")
        == b"The method of the request is not enabled by CORS rules."
    )

    for headers in ([], [(b"Origin", b"https://www.neoteroi.dev")]):
        await app(
            get_example_scope("GET", "/", headers),
            MockReceive(),
            MockSend(),
        )

        response = app.response
        assert response.status == 200
        assert response.content.body == b"Hello, World"


@pytest.mark.asyncio
async def test_cors_request_response_normalization(app):
    app.use_cors(
        allow_methods="GET POST DELETE", allow_origins="https://www.neoteroi.dev"
    )

    @app.router.get("/")
    async def home():
        return "Hello, World"

    await app.start()

    for headers in ([], [(b"Origin", b"https://www.neoteroi.dev")]):
        await app(
            get_example_scope("GET", "/", headers),
            MockReceive(),
            MockSend(),
        )

        response = app.response
        assert response.status == 200
        assert response.content.body == b"Hello, World"


@pytest.mark.asyncio
async def test_response_to_cors_request_contains_cors_headers(app):
    app.use_cors(allow_methods="GET POST", allow_origins="https://www.neoteroi.dev")

    @app.router.post("/")
    async def home():
        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "POST",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )
    assert response.headers.get_single(b"Access-Control-Expose-Headers") is not None


@pytest.mark.asyncio
async def test_response_to_failed_cors_request_contains_cors_headers(app):
    async def example(*args):
        return text("Oh, no!", status=500)

    class CrashError(Exception):
        pass

    app.exceptions_handlers[CrashError] = example

    app.use_cors(allow_methods="GET POST", allow_origins="https://www.neoteroi.dev")

    @app.router.post("/example_1")
    async def example_1():
        # example: use an exception to control the execution flow,
        # the error handler defined in app.exceptions_handlers will produce the
        # response object
        # the response object must be populated with CORS headers, otherwise the
        # client won't expose the error details - for example we might send a JSON
        # structure with error information
        raise CrashError()

    @app.router.post("/example_2")
    async def example_2():
        # example: use an exception to control the execution flow,
        # the error handler defined in app.exceptions_handlers will produce the
        # response object
        # the response object must be populated with CORS headers, otherwise the
        # client won't expose the error details - for example we might send a JSON
        # structure with error information
        raise BadRequest("Some user friendly client error detail")

    await app.start()

    await app(
        get_example_scope(
            "POST",
            "/example_1",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 500
    assert (await response.text()) == "Oh, no!"

    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )
    assert response.headers.get_single(b"Access-Control-Expose-Headers") is not None

    await app(
        get_example_scope(
            "POST",
            "/example_2",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        await response.text()
    ) == "Bad Request: Some user friendly client error detail"

    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )
    assert response.headers.get_single(b"Access-Control-Expose-Headers") is not None


@pytest.mark.asyncio
async def test_cors_preflight_request(app):
    app.use_cors(allow_methods="GET POST", allow_origins="https://www.neoteroi.dev")

    @app.router.post("/")
    async def home():
        return text("Hello, World")

    @app.router.delete("/")
    async def delete_example():
        ...

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert set(
        response.headers.get_single(b"Access-Control-Allow-Methods").split(b", ")
    ) == {b"GET", b"POST"}
    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"DELETE"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        response.headers.get_single(b"CORS-Error")
        == b"The method of the request is not enabled by CORS rules."
    )


@pytest.mark.asyncio
async def test_cors_preflight_request_allow_headers(app):
    app.use_cors(
        allow_methods="GET POST",
        allow_origins="https://www.neoteroi.dev",
        allow_headers="Authorization credentials",
    )

    @app.route("/", methods=["GET", "POST"])
    async def home():
        return text("Hello, World")

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
                (b"Access-Control-Request-Headers", b"Authorization"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert (
        response.headers.get_single(b"Access-Control-Allow-Headers") == b"Authorization"
    )

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
                (b"Access-Control-Request-Headers", b"Authorization, Credentials"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert set(
        response.headers.get_single(b"Access-Control-Allow-Headers").split(b", ")
    ) == {b"Authorization", b"Credentials"}

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
                (b"Access-Control-Request-Headers", b"X-Foo"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        response.headers.get_single(b"CORS-Error")
        == b'The "X-Foo" request header is not enabled by CORS rules.'
    )


@pytest.mark.asyncio
async def test_cors_preflight_request_allow_credentials(app):
    app.use_cors(
        allow_methods="GET POST",
        allow_origins="https://www.neoteroi.dev",
        allow_credentials=True,
    )

    @app.router.get("/")
    async def home():
        return text("Hello, World")

    @app.router.post("/")
    async def post_example():
        ...

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"Access-Control-Allow-Credentials") == b"true"


@pytest.mark.asyncio
async def test_cors_preflight_request_allow_any(app):
    app.use_cors(allow_methods="*", allow_origins="*", allow_headers="*")

    @app.router.get("/")
    async def home():
        return text("Hello, World")

    @app.router.post("/")
    async def post_example():
        ...

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"Access-Control-Allow-Methods") == b"*"
    assert response.headers.get_single(b"Access-Control-Allow-Origin") == b"*"

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
                (b"Access-Control-Request-Headers", b"X-Foo"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"Access-Control-Allow-Headers") == b"X-Foo"

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
                (b"Access-Control-Request-Headers", b"X-Ufo X-Foo"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert (
        response.headers.get_single(b"Access-Control-Allow-Headers") == b"X-Ufo X-Foo"
    )


@pytest.mark.asyncio
async def test_non_cors_options_request(app):
    app.use_cors(
        allow_methods="GET POST",
        allow_origins="https://www.neoteroi.dev",
        allow_credentials=True,
    )

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.asyncio
async def test_use_cors_raises_for_started_app(app):
    await app.start()

    with pytest.raises(ApplicationAlreadyStartedCORSError):
        app.use_cors()

    with pytest.raises(ApplicationAlreadyStartedCORSError):
        app.add_cors_policy("deny")


@pytest.mark.asyncio
async def test_add_cors_policy_configures_cors_settings(app):
    app.add_cors_policy(
        "yes", allow_methods="GET POST", allow_origins="https://www.neoteroi.dev"
    )

    @app.cors("yes")
    @app.router.post("/")
    async def home():
        return text("Hello, World")

    @app.router.post("/another")
    async def another():
        ...

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert set(
        response.headers.get_single(b"Access-Control-Allow-Methods").split(b", ")
    ) == {b"GET", b"POST"}
    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )

    await app(
        get_example_scope(
            "OPTIONS",
            "/another",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 400
    assert (
        response.headers.get_single(b"CORS-Error")
        == b"The origin of the request is not enabled by CORS rules."
    )


@pytest.mark.asyncio
async def test_cors_by_handler(app):
    app.use_cors(
        allow_methods="GET POST DELETE", allow_origins="https://www.neoteroi.dev"
    )

    app.add_cors_policy(
        "specific",
        allow_methods="GET POST",
        allow_origins="https://www.neoteroi.xyz",
    )

    @app.route("/", methods=["GET", "POST"])
    async def home():
        return text("Hello, World")

    @app.cors("specific")
    @app.route("/specific-rules", methods=["GET", "POST"])
    async def different_rules():
        return text("Specific")

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.dev"
    )
    assert set(
        response.headers.get_single(b"Access-Control-Allow-Methods").split(b", ")
    ) == {b"GET", b"POST", b"DELETE"}

    await app(
        get_example_scope(
            "OPTIONS",
            "/specific-rules",
            [
                (b"Origin", b"https://www.neoteroi.xyz"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert (
        response.headers.get_single(b"Access-Control-Allow-Origin")
        == b"https://www.neoteroi.xyz"
    )
    assert set(
        response.headers.get_single(b"Access-Control-Allow-Methods").split(b", ")
    ) == {b"GET", b"POST"}


def test_cors_decorator_raises_for_missing_policy(app):
    app.add_cors_policy(
        "yes", allow_methods="GET POST", allow_origins="https://www.neoteroi.dev"
    )

    with pytest.raises(CORSPolicyNotConfiguredError):

        @app.cors("nope")
        @app.router.post("/")
        async def home():
            return text("Hello, World")


def test_cors_decorator_raises_for_non_request_handler(app):
    app.add_cors_policy(
        "yes", allow_methods="GET POST", allow_origins="https://www.neoteroi.dev"
    )

    with pytest.raises(NotRequestHandlerError):

        @app.cors("yes")
        async def home():
            return text("Hello, World")


@pytest.mark.asyncio
async def test_cors_preflight_request_handles_404_for_missing_routes(app):
    app.use_cors(allow_methods="GET POST", allow_origins="https://www.neoteroi.dev")

    await app.start()

    await app(
        get_example_scope(
            "OPTIONS",
            "/",
            [
                (b"Origin", b"https://www.neoteroi.dev"),
                (b"Access-Control-Request-Method", b"POST"),
            ],
        ),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404



from datetime import datetime
from enum import IntEnum
from typing import Optional, Union

class CookieSameSiteMode(IntEnum):
    UNDEFINED = 0
    LAX = 1
    STRICT = 2
    NONE = 3

class CookieError(Exception): ...
class CookieValueExceedsMaximumLength(CookieError): ...

def datetime_to_cookie_format(value: datetime) -> bytes: ...
def datetime_from_cookie_format(value: bytes) -> datetime: ...

class Cookie:
    def __init__(
        self,
        name: str,
        value: str,
        expires: Optional[datetime] = None,
        domain: Optional[str] = None,
        path: Optional[str] = None,
        http_only: bool = False,
        secure: bool = False,
        max_age: int = -1,
        same_site: CookieSameSiteMode = CookieSameSiteMode.UNDEFINED,
    ):
        self.name = name
        self.value = value
        self.expires = expires
        self.domain = domain
        self.path = path
        self.http_only = http_only
        self.secure = secure
        self.max_age = max_age
        self.same_site = same_site
    def clone(self) -> "Cookie": ...
    def __eq__(self, other: Union[str, bytes, "Cookie"]) -> bool: ...
    def __repr__(self) -> str:
        return f"<Cookie {self.name}: {self.value}>"

def parse_cookie(value: bytes) -> Cookie: ...
def write_response_cookie(cookie: Cookie) -> bytes: ...
